# -*- coding: utf-8 -*-
# rdesignerProtos.py ---
#
# Filename: rdesignerProtos.py
# Description:
# Author: Subhasis Ray, Upi Bhalla
# Maintainer:
# Created: Tue May  7 12:11:22 2013 (+0530)
# Version:
# Last-Updated: Wed Dec 30 13:01:00 2015 (+0530)
#           By: Upi
# URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change log:
#
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA.
#
#

# Code:
import numpy as np
import moose
import math
from moose import utils

EREST_ACT    = -0.060
EREST_ACT_HH = -0.070 # A different value is used for the HH squid params
ECA          = 0.080
EK           = -0.075
SOMA_A       = 3.32e-9
per_ms       = 1e3
PI           = 3.14159265359
FaradayConst = 96485.3365 # Coulomb/mol
#CA_SCALE     = 25000 # Ratio of Traub units to mM. 250::0.01
CA_SCALE     = 1.0 # I have now set sensible ranges in the KCA and KAHP



def make_HH_Na(name = 'HH_Na', parent='/library', vmin=-110e-3, vmax=50e-3, vdivs=3000):
    """Create a Hodhkin-Huxley Na channel under `parent`.

    vmin, vmax, vdivs: voltage range and number of divisions for gate tables

    """
    na = moose.HHChannel('%s/%s' % (parent, name))
    na.Ek = 50e-3
    na.Xpower = 3
    na.Ypower = 1
    v = np.linspace(vmin, vmax, vdivs+1) - EREST_ACT_HH
    m_alpha = per_ms * (25 - v * 1e3) / (10 * (np.exp((25 - v * 1e3) / 10) - 1))
    m_beta = per_ms * 4 * np.exp(- v * 1e3/ 18)
    m_gate = moose.element('%s/gateX' % (na.path))
    m_gate.min = vmin
    m_gate.max = vmax
    m_gate.divs = vdivs
    m_gate.tableA = m_alpha
    m_gate.tableB = m_alpha + m_beta
    h_alpha = per_ms * 0.07 * np.exp(-v / 20e-3)
    h_beta = per_ms * 1/(np.exp((30e-3 - v) / 10e-3) + 1)
    h_gate = moose.element('%s/gateY' % (na.path))
    h_gate.min = vmin
    h_gate.max = vmax
    h_gate.divs = vdivs
    h_gate.tableA = h_alpha
    h_gate.tableB = h_alpha + h_beta
    na.tick = -1
    return na

def make_HH_K(name = 'HH_K', parent='/library', vmin=-120e-3, vmax=40e-3, vdivs=3000):
    """Create a Hodhkin-Huxley K channel under `parent`.

    vmin, vmax, vdivs: voltage range and number of divisions for gate tables

    """
    k = moose.HHChannel('%s/%s' % (parent, name))
    k.Ek = -77e-3
    k.Xpower = 4
    v = np.linspace(vmin, vmax, vdivs+1) - EREST_ACT_HH
    n_alpha = per_ms * (10 - v * 1e3)/(100 * (np.exp((10 - v * 1e3)/10) - 1))
    n_beta = per_ms * 0.125 * np.exp(- v * 1e3 / 80)
    n_gate = moose.element('%s/gateX' % (k.path))
    n_gate.min = vmin
    n_gate.max = vmax
    n_gate.divs = vdivs
    n_gate.tableA = n_alpha
    n_gate.tableB = n_alpha + n_beta
    k.tick = -1
    return k

#/========================================================================
#/                Tabchannel Na Hippocampal cell channel
#/========================================================================
def make_Na( name ):
    if moose.exists( '/library/' + name ):
        return
    Na = moose.HHChannel( '/library/' + name )
    Na.Ek = 0.055             #    V
    Na.Gbar = 300 * SOMA_A    #    S
    Na.Gk = 0                 #    S
    Na.Xpower = 2
    Na.Ypower = 1
    Na.Zpower = 0

    xgate = moose.element( Na.path + '/gateX' )
    xA = np.array( [ 320e3 * (0.0131 + EREST_ACT),
        -320e3, -1.0, -1.0 * (0.0131 + EREST_ACT), -0.004, 
        -280e3 * (0.0401 + EREST_ACT), 280e3, -1.0, 
        -1.0 * (0.0401 + EREST_ACT), 5.0e-3, 
        3000, -0.1, 0.05 ] )
    xgate.alphaParms = xA

    ygate = moose.element( Na.path + '/gateY' )
    yA = np.array( [ 128.0, 0.0, 0.0, -1.0 * (0.017 + EREST_ACT), 0.018,
        4.0e3, 0.0, 1.0, -1.0 * (0.040 + EREST_ACT), -5.0e-3, 
        3000, -0.1, 0.05 ] )
    ygate.alphaParms = yA
    return Na

#========================================================================
#                Tabchannel K(DR) Hippocampal cell channel
#========================================================================
def make_K_DR( name ):
    if moose.exists( '/library/' + name ):
        return
    K_DR = moose.HHChannel( '/library/' + name )
    K_DR.Ek = EK                #    V
    K_DR.Gbar = 150 * SOMA_A    #    S
    K_DR.Gk = 0                 #    S
    K_DR.Xpower = 1
    K_DR.Ypower = 0
    K_DR.Zpower = 0

    xgate = moose.element( K_DR.path + '/gateX' )
    xA = np.array( [ 16e3 * (0.0351 + EREST_ACT), 
        -16e3, -1.0, -1.0 * (0.0351 + EREST_ACT), -0.005,
        250, 0.0, 0.0, -1.0 * (0.02 + EREST_ACT), 0.04,
        3000, -0.1, 0.05 ] )
    xgate.alphaParms = xA
    return K_DR

#========================================================================
#                Tabchannel K(A) Hippocampal cell channel
#========================================================================
def make_K_A( name ):
    if moose.exists( '/library/' + name ):
        return
    K_A = moose.HHChannel( '/library/' + name )
    K_A.Ek = EK                #    V
    K_A.Gbar = 50 * SOMA_A    #    S
    K_A.Gk = 0                #    S
    K_A.Xpower = 1
    K_A.Ypower = 1
    K_A.Zpower = 0

    xgate = moose.element( K_A.path + '/gateX' )
    xA = np.array( [ 20e3 * (0.0131 + EREST_ACT), 
        -20e3, -1.0, -1.0 * (0.0131 + EREST_ACT), -0.01,
        -17.5e3 * (0.0401 + EREST_ACT), 
        17.5e3, -1.0, -1.0 * (0.0401 + EREST_ACT), 0.01,
        3000, -0.1, 0.05 ] )
    xgate.alphaParms = xA

    ygate = moose.element( K_A.path + '/gateY' )
    yA = np.array( [ 1.6, 0.0, 0.0, 0.013 - EREST_ACT, 0.018,
        50.0, 0.0, 1.0, -1.0 * (0.0101 + EREST_ACT), -0.005,
        3000, -0.1, 0.05 ] )
    ygate.alphaParms = yA

#//========================================================================
#//                      Tabulated Ca Channel
#//========================================================================

def make_Ca( name ):
    if moose.exists( '/library/' + name):
        return
    Ca = moose.HHChannel( '/library/' + name )
    Ca.Ek = ECA
    Ca.Gbar = 40 * SOMA_A
    Ca.Gk = 0
    Ca.Xpower = 2
    Ca.Ypower = 1
    Ca.Zpower = 0

    xgate = moose.element( Ca.path + '/gateX' )
    xA = np.array( [ 1.6e3, 0, 1.0, -1.0 * (0.065 + EREST_ACT), -0.01389, -20e3 * (0.0511 + EREST_ACT), 20e3, -1.0, -1.0 * (0.0511 + EREST_ACT), 5.0e-3, 3000, -0.1, 0.05 ] )
#    xgate.min = -0.1
#    xgate.max = 0.05
#    xgate.divs = 3000
#// Converting Traub's expressions for the gCa/s alpha and beta functions
#// to SI units and entering the A, B, C, D and F parameters, we get:
#    xgate.alpha( 1.6e3, 0, 1.0, -1.0 * (0.065 + EREST_ACT), -0.01389 )
#    xgate.beta( -20e3 * (0.0511 + EREST_ACT), 20e3, -1.0, -1.0 * (0.0511 + EREST_ACT), 5.0e-3 )
    #xgate.setupAlpha( xA )
    xgate.alphaParms = xA


#  The Y gate (gCa/r) is not quite of this form.  For V > EREST_ACT, alpha =
#  5*{exp({-50*(V - EREST_ACT)})}.  Otherwise, alpha = 5.  Over the entire
#  range, alpha + beta = 5.  To create the Y_A and Y_B tables, we use some
#  of the pieces of the setupalpha function.
    ygate = moose.element( Ca.path + '/gateY' )
    ygate.min = -0.1
    ygate.max = 0.05
    ygate.divs = 3000
    yA = np.zeros( (ygate.divs + 1), dtype=float)
    yB = np.zeros( (ygate.divs + 1), dtype=float)


#Fill the Y_A table with alpha values and the Y_B table with (alpha+beta)
    dx = (ygate.max - ygate.min)/ygate.divs
    x = ygate.min
    for i in range( ygate.divs + 1 ):
        if ( x > EREST_ACT):
            yA[i] = 5.0 * math.exp( -50 * (x - EREST_ACT) )
        else:
            yA[i] = 5.0
        yB[i] = 5.0
        x += dx
    ygate.tableA = yA
    ygate.tableB = yB
# Tell the cell reader that the current from this channel must be fed into
# the Ca_conc pool of calcium.
    addmsg1 = moose.Mstring( Ca.path + '/addmsg1' )
    addmsg1.value = '.    IkOut    ../Ca_conc    current'
# in some compartments, whe have an NMDA_Ca_conc object to put the current
# into.
    addmsg2 = moose.Mstring( Ca.path + '/addmsg2' )
    addmsg2.value = '.    IkOut    ../NMDA_Ca_conc    current'
# As we typically use the cell reader to create copies of these prototype
#elements in one or more compartments, we need some way to be sure that the
#needed messages are established.  Although the cell reader has enough
#information to create the messages which link compartments to their channels
#and to other adjacent compartments, it most be provided with the information
#needed to establish additional messages.  This is done by placing the
#message string in a user-defined field of one of the elements which is
#involved in the message.  The cell reader recognizes the added object names
#"addmsg1", "addmsg2", etc. as indicating that they are to be
#evaluated and used to set up messages.  The paths are relative to the
#element which contains the message string in its added field.  Thus,
#"../Ca_conc" refers to the sibling element Ca_conc and "."
#refers to the Ca element itself.


#/*************************************************************************
#Next, we need an element to take the Calcium current calculated by the Ca
#channel and convert it to the Ca concentration.  The "Ca_concen" object
#solves the equation dC/dt = B*I_Ca - C/tau, and sets Ca = Ca_base + C.  As
#it is easy to make mistakes in units when using this Calcium diffusion
#equation, the units used here merit some discussion.

#With Ca_base = 0, this corresponds to Traub's diffusion equation for
#concentration, except that the sign of the current term here is positive, as
#GENESIS uses the convention that I_Ca is the current flowing INTO the
#compartment through the channel.  In SI units, the concentration is usually
#expressed in moles/m^3 (which equals millimoles/liter), and the units of B
#are chosen so that B = 1/(ion_charge * Faraday * volume). Current is
#expressed in amperes and one Faraday = 96487 coulombs.  However, in this
#case, Traub expresses the concentration in arbitrary units, current in
#microamps and uses tau = 13.33 msec.  If we use the same concentration units,
#but express current in amperes and tau in seconds, our B constant is then
#10^12 times the constant (called "phi") used in the paper.  The actual value
#used will be typically be determined by the cell reader from the cell
#parameter file.  However, for the prototype channel we wlll use Traub's
#corrected value for the soma.  (An error in the paper gives it as 17,402
#rather than 17.402.)  In our units, this will be 17.402e12.

#*************************************************************************/


#//========================================================================
#//                      Ca conc
#//========================================================================

def make_Ca_conc( name ):
    if moose.exists( '/library/' + name ):
        return
    conc = moose.CaConc( '/library/tempName' )
    conc.name = name
    conc.tau = 0.013333  # sec
    conc.B  = 17.402e12 # Curr to conc conversion for soma
    conc.Ca_base = 0.00000

#This Ca_concen element should receive a message from any calcium channels
# with the current going through the channel. Here we have this specified
# in the Ca channel, with the idea that more than one channel might
# contribute Ca ions to this calcium pool. In the original GENESIS file
# this was specified here in make_Ca_conc.

#========================================================================
#             Tabulated Ca-dependent K AHP Channel: Traub 1991
#========================================================================

# This is a tabchannel which gets the calcium concentration from Ca_conc
#  in order to calculate the activation of its Z gate.  It is set up much
#  like the Ca channel, except that the A and B tables have values which are
#  functions of concentration, instead of voltage.
# Traub's original equation is min(0.2e-4 Xi, 0.01) which suggests a max
# of 500 for the Xi (calcium) in his system. For dendritic calcium say 10uM
# or 0.01 mM. Elsewhere we estimate that the ratio of Traub units to mM is
# 250:0.01. Using this ratio, we should have xmax = 0.02.

def make_K_AHP( name ):
    if moose.exists( '/library/' + name ):
        return
    K_AHP = moose.HHChannel( '/library/' + name )
    K_AHP.Ek = EK    #            V
    K_AHP.Gbar = 8 * SOMA_A #    S
    K_AHP.Gk = 0    #    S
    K_AHP.Xpower = 0
    K_AHP.Ypower = 0
    K_AHP.Zpower = 1
    K_AHP.useConcentration = 1

    zgate = moose.element( K_AHP.path + '/gateZ' )
    xmax = 0.02 # 20 micromolar.
    zgate.min = 0
    zgate.max = xmax
    zgate.divs = 3000
    zA = np.zeros( (zgate.divs + 1), dtype=float)
    zB = np.zeros( (zgate.divs + 1), dtype=float)
    dx = (zgate.max - zgate.min)/zgate.divs
    x = zgate.min
    for i in range( zgate.divs + 1 ):
            zA[i] = min( 250.00 * CA_SCALE * x, 10 )
            zB[i] = 1.0 + zA[i]
            x = x + dx

    zgate.tableA = zA
    zgate.tableB = zB
    addmsg1 = moose.Mstring( K_AHP.path + '/addmsg1' )
    addmsg1.value = '../Ca_conc    concOut    . concen'
# Use an added field to tell the cell reader to set up a message from the
# Ca_Conc with concentration info, to the current K_AHP object.


#//========================================================================
#//  Ca-dependent K Channel - K(C) - (vdep_channel with table and tabgate)
#//========================================================================

#The expression for the conductance of the potassium C-current channel has a
#typical voltage and time dependent activation gate, where the time dependence
#arises from the solution of a differential equation containing the rate
#parameters alpha and beta.  It is multiplied by a function of calcium
#concentration that is given explicitly rather than being obtained from a
#differential equation.  Therefore, we need a way to multiply the activation
#by a concentration dependent value which is determined from a lookup table.
#This is accomplished by using the Z gate with the new tabchannel "instant"
#field, introduced in GENESIS 2.2, to implement an "instantaneous" gate for
#the multiplicative Ca-dependent factor in the conductance.

def make_K_C( name ):
    if moose.exists( '/library/' + name ):
        return
    K_C = moose.HHChannel( '/library/' + name )
    K_C.Ek = EK                    #    V
    K_C.Gbar = 100.0 * SOMA_A     #    S
    K_C.Gk = 0                    #    S
    K_C.Xpower = 1
    K_C.Zpower = 1
    K_C.instant = 4                # Flag: 0x100 means Z gate is instant.
    K_C.useConcentration = 1

    # Now make a X-table for the voltage-dependent activation parameter.
    xgate = moose.element( K_C.path + '/gateX' )
    xgate.min = -0.1
    xgate.max = 0.05
    xgate.divs = 3000
    xA = np.zeros( (xgate.divs + 1), dtype=float)
    xB = np.zeros( (xgate.divs + 1), dtype=float)
    dx = (xgate.max - xgate.min)/xgate.divs
    x = xgate.min
    for i in range( xgate.divs + 1 ):
        alpha = 0.0
        beta = 0.0
        if (x < EREST_ACT + 0.05):
            alpha = math.exp( 53.872 * (x - EREST_ACT) - 0.66835 ) / 0.018975
            beta = 2000* (math.exp ( (EREST_ACT + 0.0065 - x)/0.027)) - alpha
        else:
            alpha = 2000 * math.exp( ( EREST_ACT + 0.0065 - x)/0.027 )
            beta = 0.0
        xA[i] = alpha
        xB[i] = alpha + beta
        x = x + dx
    xgate.tableA = xA
    xgate.tableB = xB

# Create a table for the function of concentration, allowing a
# concentration range of 0 to 100 uM, with 3000 divisions.  This is done
# using the Z gate, which can receive a CONCEN message.  By using
# the "instant" flag, the A and B tables are evaluated as lookup tables,
#  rather than being used in a differential equation.
    zgate = moose.element( K_C.path + '/gateZ' )
    zgate.min = 0.0
    xmax = 0.02  # Max Ca conc is likely to be 100 uM even in spine, but
        # based on estimates above let's keep it at 20uM.
    zgate.max = xmax
    zgate.divs = 3000
    zA = np.zeros( (zgate.divs + 1), dtype=float)
    zB = np.zeros( (zgate.divs + 1), dtype=float)
    dx = ( zgate.max -  zgate.min)/ zgate.divs
    x = zgate.min
    #CaScale = 100000.0 / 250.0e-3
    for i in range( zgate.divs + 1 ):
        zA[i] = min( 1.0, x * CA_SCALE / xmax)
        zB[i] = 1.0
        x += dx
    zgate.tableA = zA
    zgate.tableB = zB

# Now we need to provide for messages that link to external elements.
# The message that sends the Ca concentration to the Z gate tables is stored
# in an added field of the channel, so that it may be found by the cell
# reader.
    addmsg1 = moose.Mstring( K_C.path + '/addmsg1' )
    addmsg1.value = '../Ca_conc    concOut    . concen'


#========================================================================
#                SynChan: Glu receptor
#========================================================================

def make_glu( name ):
    if moose.exists( '/library/' + name ):
        return
    glu = moose.SynChan( '/library/' + name )
    glu.Ek = 0.0
    glu.tau1 = 2.0e-3
    glu.tau2 = 9.0e-3
    sh = moose.SimpleSynHandler( glu.path + '/sh' )
    moose.connect( sh, 'activationOut', glu, 'activation' )
    sh.numSynapses = 1
    sh.synapse[0].weight = 1
    return glu

#========================================================================
#                NMDAChan: NMDA receptor
#========================================================================

def make_NMDA( name ):
    if moose.exists( '/library/' + name ):
        return
    NMDA = moose.NMDAChan( '/library/' + name )
    NMDA.Ek = 0.0
    NMDA.tau1 = 20.0e-3
    NMDA.tau2 = 20.0e-3
    NMDA.Gbar = 5 * SOMA_A
    NMDA.CMg = 1.2		#	[Mg]ext in mM
    NMDA.KMg_A = 1.0/0.28
    NMDA.KMg_B = 1.0/62
    NMDA.temperature = 300  # Temperature in Kelvin.
    NMDA.extCa = 1.5        # [Ca]ext in mM
    NMDA.intCa = 0.00008        # [Ca]int in mM
    NMDA.intCaScale = 1         # Scale factor from elec Ca units to mM
    NMDA.intCaOffset = 0.00008  # Basal [Ca]int in mM
    NMDA.condFraction = 0.02  # Fraction of conductance due to Ca

    addmsg1 = moose.Mstring( NMDA.path + '/addmsg1' )
    addmsg1.value = '.	ICaOut ../Ca_conc current'
    addmsg2 = moose.Mstring( NMDA.path + '/addmsg2' )
    addmsg2.value = '../Ca_conc	concOut . assignIntCa'

    sh = moose.SimpleSynHandler( NMDA.path + '/sh' )
    moose.connect( sh, 'activationOut', NMDA, 'activation' )
    sh.numSynapses = 1
    sh.synapse[0].weight = 1
    return NMDA

#========================================================================
#                SynChan: GABA receptor
#========================================================================

def make_GABA( name ):
    if moose.exists( '/library/' + name ):
        return
    GABA = moose.SynChan( '/library/' + name )
    GABA.Ek = -0.065            # V
    GABA.tau1 = 4.0e-3          # s
    GABA.tau2 = 9.0e-3          # s
    sh = moose.SimpleSynHandler( GABA.path + '/sh' )
    moose.connect( sh, 'activationOut', GABA, 'activation' )
    sh.numSynapses = 1
    sh.synapse[0].weight = 1

#========================================================================

def makeChemOscillator( name = 'osc', parent = '/library' ):
    model = moose.Neutral( parent + '/' + name )
    compt = moose.CubeMesh( model.path + '/kinetics' )
    """
    This function sets up a simple oscillatory chemical system within
    the script. The reaction system is::

        s ---a---> a  // s goes to a, catalyzed by a.
        s ---a---> b  // s goes to b, catalyzed by a.
        a ---b---> s  // a goes to s, catalyzed by b.
        b -------> s  // b is degraded irreversibly to s.

    in sum, **a** has a positive feedback onto itself and also forms **b**.
    **b** has a negative feedback onto **a**.
    Finally, the diffusion constant for **a** is 1/10 that of **b**.
    """
    # create container for model
    diffConst = 10e-12 # m^2/sec
    motorRate = 1e-6 # m/sec
    concA = 1 # millimolar

    # create molecules and reactions
    a = moose.Pool( compt.path + '/a' )
    b = moose.Pool( compt.path + '/b' )
    s = moose.Pool( compt.path + '/s' )
    e1 = moose.MMenz( compt.path + '/e1' )
    e2 = moose.MMenz( compt.path + '/e2' )
    e3 = moose.MMenz( compt.path + '/e3' )
    r1 = moose.Reac( compt.path + '/r1' )

    a.concInit = 0.1
    b.concInit = 0.1
    s.concInit = 1

    moose.connect( e1, 'sub', s, 'reac' )
    moose.connect( e1, 'prd', a, 'reac' )
    moose.connect( a, 'nOut', e1, 'enzDest' )
    e1.Km = 1
    e1.kcat = 1

    moose.connect( e2, 'sub', s, 'reac' )
    moose.connect( e2, 'prd', b, 'reac' )
    moose.connect( a, 'nOut', e2, 'enzDest' )
    e2.Km = 1
    e2.kcat = 0.5

    moose.connect( e3, 'sub', a, 'reac' )
    moose.connect( e3, 'prd', s, 'reac' )
    moose.connect( b, 'nOut', e3, 'enzDest' )
    e3.Km = 0.1
    e3.kcat = 1

    moose.connect( r1, 'sub', b, 'reac' )
    moose.connect( r1, 'prd', s, 'reac' )
    r1.Kf = 0.3 # 1/sec
    r1.Kb = 0 # 1/sec

    # Assign parameters
    a.diffConst = diffConst/10
    b.diffConst = diffConst
    s.diffConst = 0
    return compt


    #################################################################
    # Here we have a series of utility functions for building cell
    # prototypes.
    #################################################################
def transformNMDAR( path ):
    for i in moose.wildcardFind( path + "/##/#NMDA#[ISA!=NMDAChan]" ):
        chanpath = i.path
        pa = i.parent
        i.name = '_temp'
        if ( chanpath[-3:] == "[0]" ):
            chanpath = chanpath[:-3]
        nmdar = moose.NMDAChan( chanpath )
        sh = moose.SimpleSynHandler( chanpath + '/sh' )
        moose.connect( sh, 'activationOut', nmdar, 'activation' )
        sh.numSynapses = 1
        sh.synapse[0].weight = 1
        nmdar.Ek = i.Ek
        nmdar.tau1 = i.tau1
        nmdar.tau2 = i.tau2
        nmdar.Gbar = i.Gbar
        nmdar.CMg = 12
        nmdar.KMg_A = 1.0 / 0.28
        nmdar.KMg_B = 1.0 / 62
        nmdar.temperature = 300
        nmdar.extCa = 1.5
        nmdar.intCa = 0.00008
        nmdar.intCaScale = 1
        nmdar.intCaOffset = 0.00008
        nmdar.condFraction = 0.02
        moose.delete( i )
        moose.connect( pa, 'channel', nmdar, 'channel' )
        caconc = moose.wildcardFind( pa.path + '/#[ISA=CaConcBase]' )
        if ( len( caconc ) < 1 ):
            print('no caconcs found on ', pa.path)
        else:
            moose.connect( nmdar, 'ICaOut', caconc[0], 'current' )
            moose.connect( caconc[0], 'concOut', nmdar, 'assignIntCa' )
    ################################################################
    # Utility function for building a compartment, used for spines.
    # Builds a compartment object downstream (further away from soma)
    # of the specfied previous compartment 'pa'. If 'pa' is not a
    # compartment, it builds it on 'pa'. It places the compartment
    # on the end of 'prev', and at 0,0,0 otherwise.

def buildCompt( pa, name, RM = 1.0, RA = 1.0, CM = 0.01, dia = 1.0e-6, x = 0.0, y = 0.0, z = 0.0, dx = 10e-6, dy = 0.0, dz = 0.0, Em = -0.065, initVm = -0.065 ):
    length = np.sqrt( dx * dx + dy * dy + dz * dz )
    compt = moose.Compartment( pa.path + '/' + name )
    compt.x0 = x
    compt.y0 = y
    compt.z0 = z
    compt.x = dx + x
    compt.y = dy + y
    compt.z = dz + z
    compt.diameter = dia
    compt.length = length
    xa = dia * dia * PI / 4.0
    sa = length * dia * PI
    compt.Ra = length * RA / xa
    compt.Rm = RM / sa
    compt.Cm = CM * sa
    compt.Em = Em
    compt.initVm = initVm
    return compt

def buildComptWrapper( pa, name, length, dia, xoffset, RM, RA, CM ):
    return buildCompt( pa, name, RM, RA, CM, dia = dia, x = xoffset, dx = length )

    ################################################################
    # Utility function for building a synapse, used for spines.
def buildSyn( name, compt, Ek, tau1, tau2, Gbar, CM ):
    syn = moose.SynChan( compt.path + '/' + name )
    syn.Ek = Ek
    syn.tau1 = tau1
    syn.tau2 = tau2
    syn.Gbar = Gbar * compt.Cm / CM
    #print "BUILD SYN: ", name, Gbar, syn.Gbar, CM
    moose.connect( compt, 'channel', syn, 'channel' )
    sh = moose.SimpleSynHandler( syn.path + '/sh' )
    moose.connect( sh, 'activationOut', syn, 'activation' )
    sh.numSynapses = 1
    sh.synapse[0].weight = 1
    return syn

######################################################################
# Utility function, borrowed from proto18.py, for making an LCa channel.
# Based on Traub's 91 model, I believe.
def make_LCa( name = 'LCa', parent = '/library' ):
        if moose.exists( parent + '/' + name ):
                return
        Ca = moose.HHChannel( parent + '/' + name )
        Ca.Ek = ECA
        Ca.Gbar = 0
        Ca.Gk = 0
        Ca.Xpower = 2
        Ca.Ypower = 1
        Ca.Zpower = 0

        xgate = moose.element( parent + '/' + name + '/gateX' )
        xA = np.array( [ 1.6e3, 0, 1.0, -1.0 * (0.065 + EREST_ACT), -0.01389, -20e3 * (0.0511 + EREST_ACT), 20e3, -1.0, -1.0 * (0.0511 + EREST_ACT), 5.0e-3, 3000, -0.1, 0.05 ] )
        xgate.alphaParms = xA
        ygate = moose.element( parent + '/' + name + '/gateY' )
        ygate.min = -0.1
        ygate.max = 0.05
        ygate.divs = 3000
        yA = np.zeros( (ygate.divs + 1), dtype=float)
        yB = np.zeros( (ygate.divs + 1), dtype=float)


#Fill the Y_A table with alpha values and the Y_B table with (alpha+beta)
        dx = (ygate.max - ygate.min)/ygate.divs
        x = ygate.min
        for i in range( ygate.divs + 1 ):
                if ( x > EREST_ACT):
                        yA[i] = 5.0 * math.exp( -50 * (x - EREST_ACT) )
                else:
                        yA[i] = 5.0
                yB[i] = 5.0
                x += dx
        ygate.tableA = yA
        ygate.tableB = yB
        return Ca
######################################################################

def make_vclamp( name = 'Vclamp', parent = '/library' ):
    if moose.exists( '/library/' + name ):
        return
    vclamp = moose.VClamp( parent + '/' + name )
    vclamp.mode = 0     # Default. could try 1, 2 as well
    vclamp.tau = 0.2e-3 # lowpass filter for command voltage input
    vclamp.ti = 20e-6   # Integral time
    vclamp.td = 5e-6    # Differential time. Should it be >= dt?
    vclamp.gain = 0.00005   # Gain of vclamp ckt.

    # Connect voltage clamp circuitry
    addmsg1 = moose.Mstring( vclamp.path + '/addmsg1' )
    addmsg1.value = '.  currentOut  ..  injectMsg'
    addmsg2 = moose.Mstring( vclamp.path + '/addmsg2' )
    addmsg2.value = '.. VmOut . sensedIn'

    return vclamp

######################################################################

def make_synInput( name = 'RandSpike', parent = '/library' ):
    if moose.exists( '/library/' + name ):
        return
    rs = moose.RandSpike( parent + '/' + name + '_rs' )
    rs.rate = 0     # mean firing rate
    rs.refractT = 5e-3 # 5 ms.
    

    # Connect rand spike to channel that it is sitting on.
    addmsg1 = moose.Mstring( rs.path + '/addmsg1' )
    addmsg1.value = '.  spikeOut  ../sh/synapse[0]  addSpike'

    return rs

######################################################################

    ################################################################
    # API function for building spine prototypes. Here we put in the
    # spine dimensions, and options for standard channel types.
    # The synList tells it to create dual alpha function synchans:
    # [name, Erev, tau1, tau2, conductance_density, connectToCa]
    # The chanList tells it to copy over channels defined in /library
    # and assign the specified conductance density.
    # If caTau <= zero then there is no caConc created, otherwise it
    # creates one and assigns the desired tau in seconds.
    # With the default arguments here it will create a glu, NMDA and LCa,
    # and add a Ca_conc.
def addSpineProto( name = 'spine',
        parent = '/library',
        RM = 1.0, RA = 1.0, CM = 0.01,
        shaftLen = 1.e-6 , shaftDia = 0.2e-6,
        headLen = 0.5e-6, headDia = 0.5e-6,
        synList = (),
        chanList = (),
        caTau = 0.0
        ):
    assert( moose.exists( parent ) ), "%s must exist" % parent
    spine = moose.Neutral( parent + '/' + name )
    shaft = buildComptWrapper( spine, 'shaft', shaftLen, shaftDia, 0.0, RM, RA, CM )
    head = buildComptWrapper( spine, 'head', headLen, headDia, shaftLen, RM, RA, CM )
    moose.connect( shaft, 'axial', head, 'raxial' )

    if caTau > 0.0:
        conc = moose.CaConc( head.path + '/Ca_conc' )
        conc.tau = caTau
        conc.length = head.length
        conc.diameter = head.diameter
        conc.thick = 0.0
        # The 'B' field is deprecated.
        # B = 1/(ion_charge * Faraday * volume)
        #vol = head.length * head.diameter * head.diameter * PI / 4.0
        #conc.B = 1.0 / ( 2.0 * FaradayConst * vol )
        conc.Ca_base = 0.0
    for i in synList:
        syn = buildSyn( i[0], head, i[1], i[2], i[3], i[4], CM )
        if i[5] and caTau > 0.0:
            moose.connect( syn, 'IkOut', conc, 'current' )
    for i in chanList:
        if ( moose.exists( parent + '/' + i[0] ) ):
            chan = moose.copy( parent + '/' + i[0], head )
        else:
            moose.setCwe( head )
            chan = make_LCa()
            chan.name = i[0]
            moose.setCwe( '/' )
        chan.Gbar = i[1] * head.Cm / CM
        #print "CHAN = ", chan, chan.tick, chan.Gbar
        moose.connect( head, 'channel', chan, 'channel' )
        if i[2] and caTau > 0.0:
            moose.connect( chan, 'IkOut', conc, 'current' )
    transformNMDAR( parent + '/' + name )
    return spine

#######################################################################
# Here are some compartment related prototyping functions
def makePassiveHHsoma(name = 'passiveHHsoma', parent='/library'):
    ''' Make HH squid model sized compartment:
    len and dia 500 microns. CM = 0.01 F/m^2, RA =
    '''
    elecpath = parent + '/' + name
    if not moose.exists( elecpath ):
        elecid = moose.Neuron( elecpath )
        dia = 500e-6
        soma = buildCompt( elecid, 'soma', dx = dia, dia = dia, x = 0.0,
            RM = 0.33333333, RA = 3000, CM = 0.01 )
        soma.initVm = -65e-3 # Resting of -65, from HH
        soma.Em = -54.4e-3 # 10.6 mV above resting of -65, from HH
    else:
        elecid = moose.element( elecpath )
    return elecid

# Wrapper function. This is used by the proto builder from rdesigneur
def makeActiveSpine(name = 'active_spine', parent='/library'):
    return addSpineProto( name = name, parent = parent,
            synList = ( ['glu', 0.0, 2e-3, 9e-3, 200.0, False],
            ['NMDA', 0.0, 20e-3, 20e-3, 80.0, True] ),
            chanList = ( ['Ca', 10.0, True ], ),
            caTau = 13.333e-3
        )

# Wrapper function. This is used by the proto builder from rdesigneur
def makeExcSpine(name = 'exc_spine', parent='/library'):
    return addSpineProto( name = name, parent = parent,
            synList = ( ['glu', 0.0, 2e-3, 9e-3, 200.0, False],
            ['NMDA', 0.0, 20e-3, 20e-3, 80.0, True] ),
            caTau = 13.333e-3 )

# Wrapper function. This is used by the proto builder from rdesigneur
def makePassiveSpine(name = 'passive_spine', parent='/library'):
    return addSpineProto( name = name, parent = parent)


# legacy function. This is used by the proto builder from rdesigneur
def makeSpineProto( name ):
    addSpineProto( name = name, chanList = () )



"""test_rdesigneur.py: 

"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2017-, Dilawar Singh"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import sys
import os
import numpy as np
import moose
import rdesigneur as rd 

def test_rdes2():
    if moose.exists( '/model' ):
        moose.delete( '/model' )

    rdes = rd.rdesigneur(
        stimList = [['soma', '1', '.', 'inject', '(t>0.1 && t<0.2) * 2e-8' ]],
        plotList = [['soma', '1', '.', 'Vm', 'Soma membrane potential']]
    )
    rdes.buildModel()
    moose.reinit()
    moose.start( 0.3 )

def test_rdes1():
    rdes = rd.rdesigneur()
    rdes.buildModel()
    moose.showfield(rdes.soma)

def main( ):
    test_rdes1()
    test_rdes2()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import moose
import re

def fixPath(path):
    path = re.sub(r'/+', '/', path)
    return path

def test_existing_path():
    paths = [ '/a'
            , '//a'
            , '/a/b'
            , '/a/b/'
            , '//a//b/////'
            , '/a/./b'
            , '///a/././b'
            ]
    expectedPath = [fixPath(p) for p in paths]

    expectedChanges = zip(paths, expectedPath)

    for p in expectedPath:
        print( 'Accessing %s' % p )
        try:
            moose.Neutral(p)
        except Exception as e:
            print( 'Same path access by element' )
            moose.element( p )

    foundPath = []
    for p in moose.wildcardFind('/##'):
        if "/a" in p.path:
           foundPath.append(p.path)

    testFailed = False
    for f in foundPath:
        f = re.sub(r'\[\d+\]', '', f)
        if f not in expectedPath:
            testFailed = True

    assert not testFailed, "Test failed on paths"

def main():
    test_existing_path()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#*******************************************************************
# * File:            merge.py
# * Description:
# * Author:          HarshaRani
# * E-mail:          hrani@ncbs.res.in
# ********************************************************************/
# **********************************************************************
#** This program is part of 'MOOSE', the
#** Messaging Object Oriented Simulation Environment,
#** also known as GENESIS 3 base code.
#**           copyright (C) 2003-2017 Upinder S. Bhalla. and NCBS
#Created : Friday Dec 16 23:19:00 2016(+0530)
#Version
#Last-Updated: Wed May 21 11:52:33 2018(+0530)
#         By: Harsha
#**********************************************************************/

# This program is used to merge chem models from src to destination
#Rules are :
#   -- If Compartment from the src model doesn't exist in destination model,
#       then entire compartment and its children are copied over including groups
#   -- Models are mergered at group level (if exists)
#       (Group is Neutral object in moose, which may represent pathway in network model)
#   -- Pool's are copied from source to destination if it doesn't exist, if exist nothing is done
#   -- Reaction (Reac), Enzyme (Enz) are copied
#       --- if any danglling Reac or Enz exist then that is not copied
#
#       --- if Reac Name's is different for a given path (group level)
#            then copy the entire Reac along with substrate/product
#       --- if same Reac Name and same sub and prd then nothing is copied
#       --- if same Reac Name but sub or prd is different then duplicated and copied
#
#       --- if Enz Name's is different for a given parent pool path
#            then copy the entire Enz along with substrate/product
#       --- if same Enz Name and same sub and prd then nothing is copied
#       --- if same Enz Name but sub or prd is different then duplicated and copied
#   -- Function are copied only if destination pool to which its suppose to connect doesn't exist with function of its own
#       which is a limitation in moose that no function can be connected to same pool
#
'''
Change log
May 21 : Instead of A and B changed to S and D (source and destination), 
        - at Pool level Neutral and its objected where copied, but again at Enz and Reac level copying was done which caused duplicates which is taken care
        - If Neutral object name is changed for destination, then change made to source file which would be easy to copy else path issue will come
Oct 25 : line to load SBML file was commented, which is uncommented now and also while copying MMenz had a problem which also cleaned up
Oct 14 : absolute import with mtypes just for python3

Oct 12 : clean way of cheking the type of path provided, filepath,moose obj, moose path are taken,
        if source is empty then nothing to copy, 
        if destination was empty list is update with new object

Oct 11 : missing group are copied instead of creating one in new path which also copies Annotator info
        earlier if user asked to save the model, it was saving default to kkit format, now user need to run the command to save (if this is run in command)
        To check: When Gui is allowed to merge 2 models, need to see what happens

'''

import sys
import os
#from . import _moose as moose
import moose

from moose.chemMerge import mtypes
from moose.chemUtil.chemConnectUtil import *
from moose.chemUtil.graphUtils import *
#from moose.genesis import mooseWriteKkit

def checkFile_Obj_str(file_Obj_str):
    model = moose.element('/')
    loaded = False
    found = False
    if isinstance(file_Obj_str, str):
        if os.path.isfile(file_Obj_str) == True:
            model,loaded = loadModels(file_Obj_str)
            found = True
        elif file_Obj_str.find('/') != -1 :
            if not isinstance(file_Obj_str,moose.Neutral):
                if moose.exists(file_Obj_str):
                    model = file_Obj_str
                    loaded = True
                    found = True
        elif isinstance(file_Obj_str, moose.Neutral):
            if moose.exists(file_Obj_str.path):
                model = file_Obj_str.path
                loaded = True
                found = True
    elif isinstance(file_Obj_str, moose.Neutral):
        if moose.exists(file_Obj_str.path):
            model = file_Obj_str.path
            loaded = True
            found = True
    if not found:
        print ("%s path or filename doesnot exist. " % (file_Obj_str))
    return model,loaded

def mergeChemModel(src,des):
    """ Merges two model or the path """
    A = src
    B = des
    sfile = src
    dfile = des
    loadedA = False
    loadedB = False
    modelA = moose.element('/')
    modelB = moose.element('/')
    modelA,loadedA = checkFile_Obj_str(A)
    modelB,loadedB = checkFile_Obj_str(B)
    
    if loadedA and loadedB:
        ## yet deleteSolver is called to make sure all the moose object are off from solver
        deleteSolver(modelA)
        deleteSolver(modelB)

        global poolListina
        poolListina = {}
        grpNotcopiedyet = []
        dictComptA = dict( [ (i.name,i) for i in moose.wildcardFind(modelA+'/##[ISA=ChemCompt]') ] )
        dictComptB = dict( [ (i.name,i) for i in moose.wildcardFind(modelB+'/##[ISA=ChemCompt]') ] )
                
        poolNotcopiedyet = []
        if len(dictComptA):
            for key in list(dictComptA.keys()):
                if key not in dictComptB:
                    # if compartment name from modelB does not exist in modelA, then copy
                    copy = moose.copy(dictComptA[key],moose.element(modelB))

                    dictComptB[key]=moose.element(copy)
                else:
                    #if compartment name from modelB exist in modelA,
                    #volume is not same, then change volume of ModelB same as ModelA
                    if abs(dictComptB[key].volume - dictComptA[key].volume):
                        #hack for now
                        while (abs(dictComptB[key].volume - dictComptA[key].volume) != 0.0):
                            dictComptA[key].volume = float(dictComptB[key].volume)
                    dictComptB = dict( [ (i.name,i) for i in moose.wildcardFind(modelB+'/##[ISA=ChemCompt]') ] )
                    
                    #Mergering pool
                    poolMerge(dictComptA[key],dictComptB[key],poolNotcopiedyet)

            
            comptBdict =  comptList(modelB)

            poolListinb = {}
            poolListinb = updatePoolList(comptBdict)
            
            R_Duplicated, R_Notcopiedyet,R_Dangling = [], [], []
            E_Duplicated, E_Notcopiedyet,E_Dangling = [], [], []
            for key in list(dictComptA.keys()):
                funcExist, funcNotallowed = [], []
                funcExist,funcNotallowed = functionMerge(dictComptB,dictComptA,key)

                poolListinb = updatePoolList(dictComptB)
                R_Duplicated,R_Notcopiedyet,R_Dangling = reacMerge(dictComptA,dictComptB,key,poolListinb)

                poolListinb = updatePoolList(dictComptB)
                E_Duplicated,E_Notcopiedyet,E_Dangling = enzymeMerge(dictComptB,dictComptA,key,poolListinb)
            
            # if isinstance(src, str):
            #     if os.path.isfile(src) == True:
            #         spath, sfile = os.path.split(src)
            #     else:
            #         sfile = src
            # else:
            #     sfile = src
            # if isinstance(des, str):
            #     print " A str",des
            #     if os.path.isfile(des) == True:
            #         dpath, dfile = os.path.split(des)
            #     else:
            #         dfile = des
            # else:
            #     dfile = des
            
            print("\nThe content of %s (src) model is merged to %s (des)." %(sfile, dfile))
            # Here any error or warning during Merge is written it down
            if funcExist:
                print( "\nIn model \"%s\" pool already has connection from a function, these function from model \"%s\" is not allowed to connect to same pool,\n since no two function are allowed to connect to same pool:"%(dfile, sfile))
                for fl in list(funcExist):
                    print("\t [Pool]:  %s [Function]:  %s \n" %(str(fl.parent.name), str(fl.path)))
            if funcNotallowed:
                print( "\nThese functions is not to copied, since pool connected to function input are from different compartment:")
                for fl in list(funcNotallowed):
                    print("\t [Pool]:  %s [Function]:  %s \n" %(str(fl.parent.name), str(fl.path)))
            if R_Duplicated or E_Duplicated:
                print ("These Reaction / Enzyme are \"Duplicated\" into destination file \"%s\", due to "
                        "\n 1. If substrate / product name's are different for a give reaction/Enzyme name "
                        "\n 2. If product belongs to different compartment "
                        "\n Models have to decide to keep or delete these reaction/enzyme in %s" %(dfile, dfile))
                if E_Duplicated:
                    print("Reaction: ")
                for rd in list(R_Duplicated):
                    print ("%s " %str(rd.name))

                if E_Duplicated:
                    print ("Enzyme:")
                    for ed in list(E_Duplicated):
                        print ("%s " %str(ed.name))

            if R_Notcopiedyet or E_Notcopiedyet:

                print ("\nThese Reaction/Enzyme in model are not dagging but while copying the associated substrate or product is missing")
                if R_Notcopiedyet:
                    print("Reaction: ")
                for rd in list(R_Notcopiedyet):
                    print ("%s " %str(rd.name))
                if E_Notcopiedyet:
                    print ("Enzyme:")
                    for ed in list(E_Notcopiedyet):
                        print ("%s " %str(ed.name))

            if R_Dangling or E_Dangling:
                print ("\n Dangling reaction/enzyme are not allowed in moose, these are not merged to %s from %s" %(dfile, sfile))
                if R_Dangling:
                    print("Reaction: ")
                    for rd in list(R_Dangling):
                        print ("%s " %str(rd.name))
                if E_Dangling:
                    print ("Enzyme:")
                    for ed in list(E_Dangling):
                        print ("%s " %str(ed.name))
            
            ## Model is saved
            print ("\n ")
            print ('\nMerged model is available under moose.element(\'%s\')' %(modelB))
            print ('  From the python terminal itself \n to save the model in to genesis format use \n   >moose.mooseWriteKkit(\'%s\',\'filename.g\')' %(modelB))
            print ('  to save into SBML format \n   >moose.mooseWriteSBML(\'%s\',\'filename.xml\')' %(modelB))
            return modelB
            # savemodel = raw_input("Do you want to save the model?  \"YES\" \"NO\" ")
            # if savemodel.lower() == 'yes' or savemodel.lower() == 'y':
            #     mergeto = raw_input("Enter File name ")
            #     if mergeto and mergeto.strip():
            #         filenameto = 'merge.g'
            #     else:
            #         if str(mergeto).rfind('.') != -1:
            #             mergeto = mergeto[:str(mergeto).rfind('.')]
            #         if str(mergeto).rfind('/'):
            #             mergeto = mergeto+'merge'

            #         filenameto = mergeto+'.g'

            #     error,written = moose.mooseWriteKkit(modelB, filenameto)
            #     if written == False:
            #         print('Could not save the Model, check the files')
            #     else:
            #         if error == "":
            #             print(" \n The merged model is saved into \'%s\' " %(filenameto))
            #         else:
            #             print('Model is saved but these are not written\n %s' %(error))

            # else:
            #     print ('\nMerged model is available under moose.element(\'%s\')' %(modelB))
            #     print ('  If you are in python terminal you could save \n   >moose.mooseWriteKkit(\'%s\',\'filename.g\')' %(modelB))
            #     print ('  If you are in python terminal you could save \n   >moose.mooseWriteSBML(\'%s\',\'filename.g\')' %(modelB))
            #return modelB
        
    else:
        print ('\nSource file has no objects to copy(\'%s\')' %(modelA))
        return moose.element('/')
def functionMerge(comptA,comptB,key):
    funcNotallowed, funcExist = [], []
    comptApath = moose.element(comptA[key]).path
    comptBpath = moose.element(comptB[key]).path
    objA = moose.element(comptApath).parent.name
    objB = moose.element(comptBpath).parent.name

    #This will give us all the function which exist in modelB
    funcListinb = moose.wildcardFind(comptBpath+'/##[ISA=Function]')
    for fb in funcListinb:
        #This will give us all the pools that its connected to, for this function
        fvalueOut = moose.element(fb).neighbors['valueOut']
        for poolinB in fvalueOut:
            poolinBpath = poolinB.path
            poolinA = poolinBpath.replace(objB,objA)
            connectionexist = []
            if moose.exists(poolinA):
                #This is give us if pool which is to be connected already exist any connection
                connectionexist = moose.element(poolinA).neighbors['setN']+moose.element(poolinA).neighbors['setConc']+ moose.element(poolinA).neighbors['increment']
                if len(connectionexist) == 0:
                    #This pool in model A doesnot exist with any function
                    inputs = moose.element(fb.path+'/x').neighbors['input']
                    volumes = []
                    for ins in inputs:
                        volumes.append((findCompartment(moose.element(ins))).volume)
                    if len(set(volumes)) == 1:
                        # If all the input connected belongs to one compartment then copy
                        createFunction(fb,poolinA,objB,objA)
                    else:
                        # moose doesn't allow function's input to come from different compartment
                        funcNotallowed.append(fb)
                else:
                    #Pool in model 'A' already exist function "
                    funcExist.append(fb)
            else:
                print(" Path in model A doesn't exists %s" %(poolinA))

    return funcExist,funcNotallowed

def createFunction(fb,setpool,objB,objA):
    fapath1 = fb.path.replace(objB,objA)
    fapath = fapath1.replace('[0]','')

    if not moose.exists(fapath):
        # if fb.parent.className in ['CubeMesh','CyclMesh']:
        #   des = moose.Function('/'+objA+'/'+fb.parent.name+'/'+fb.name)
        # elif fb.parent.className in ['Pool','ZombiePool','BufPool','ZombieBufPool']:
        #   for akey in list(poolListina[findCompartment(fb).name]):
        #       if fb.parent.name == akey.name:
        #           des = moose.Function(akey.path+'/'+fb.name)
        des = moose.Function(fapath)
    else:
        des = moose.element(fapath)
    inputB = moose.element(fb.path+'/x').neighbors["input"]
    moose.connect(des, 'valueOut', moose.element(setpool),'setN' )
    inputA = []
    inputA = moose.element(fapath+'/x').neighbors["input"]
    if not inputA:
        for src in inputB:
            pool = ((src.path).replace(objB,objA)).replace('[0]','')
            numVariables = des.numVars
            expr = ""
            expr = (des.expr+'+'+'x'+str(numVariables))
            expr = expr.lstrip("0 +")
            expr = expr.replace(" ","")
            des.expr = expr
            moose.connect( pool, 'nOut', des.x[numVariables], 'input' )

def comptList(modelpath):
    comptdict = {}
    for ca in moose.wildcardFind(modelpath+'/##[ISA=ChemCompt]'):
        comptdict[ca.name] = ca
    return comptdict

def loadModels(filepath):
    """ load models into moose if file, if moosepath itself it passes back the path and
    delete solver if exist """

    modelpath = '/'
    loaded = False

    if os.path.isfile(filepath) :
        fpath, filename = os.path.split(filepath)
        # print " head and tail ",head,  " ",tail
        # modelpath = filename[filename.rfind('/'): filename.rfind('.')]
        # print "modelpath ",modelpath
        # ext = os.path.splitext(filename)[1]
        # filename = filename.strip()
        modelpath = '/'+filename[:filename.rfind('.')]
        modeltype = mtypes.getType(filepath)
        subtype = mtypes.getSubtype(filepath, modeltype)

        if subtype == 'kkit' or modeltype == "cspace":
            if moose.exists(modelpath):
                moose.delete(modelpath)
            moose.loadModel(filepath,modelpath)
            loaded = True

        elif subtype == 'sbml':
            if moose.exists(modelpath):
                moose.delete(modelpath)
            moose.mooseReadSBML(filepath,modelpath)
            loaded = True
        else:
            print("This file is not supported for mergering")
            modelpath = moose.Shell('/')

    elif moose.exists(filepath):
        modelpath = filepath
        loaded = True

    return modelpath,loaded

def deleteSolver(modelRoot):
    compts = moose.wildcardFind(modelRoot+'/##[ISA=ChemCompt]')
    for compt in compts:
        if moose.exists(compt.path+'/stoich'):
            st = moose.element(compt.path+'/stoich')
            st_ksolve = st.ksolve
            moose.delete(st)
            if moose.exists((st_ksolve).path):
                moose.delete(st_ksolve)

def poolMerge(comptS,comptD,poolNotcopiedyet):
    #Here from source file all the groups are check if exist, if doesn't exist then create those groups
    #Then for that group pools are copied
    SCmptGrp = moose.wildcardFind(comptS.path+'/#[TYPE=Neutral]')
    SCmptGrp = SCmptGrp +(moose.element(comptS.path),)
    
    DCmptGrp = moose.wildcardFind(comptD.path+'/#[TYPE=Neutral]')
    DCmptGrp = DCmptGrp +(moose.element(comptD.path),)
    
    objS = moose.element(comptS.path).parent.name
    objD = moose.element(comptD.path).parent.name
    
    for spath in SCmptGrp:
        grp_cmpt = ((spath.path).replace(objS,objD)).replace('[0]','')
        if moose.exists(grp_cmpt):
            #If path exist, but its not the Neutral obj then creating a neutral obj with _grp
            #It has happened that pool, reac, enz name might exist with the same name, which when tried to create a group
            # it silently ignored the path and object copied under that pool instead of Group
            if moose.element(grp_cmpt).className != spath.className:
                grp_cmpt = grp_cmpt+'_grp'
                moose.Neutral(grp_cmpt)
                # If group name is changed while creating in destination, then in source file the same is changed,
                # so that later path issue doens't come 
                spath.name = spath.name+'_grp'
        else:
            #Neutral obj from src if doesn't exist in destination,then create src's Neutral obj in des
            src = spath
            srcpath = (spath.parent).path
            des = srcpath.replace(objS,objD)
            moose.Neutral(moose.element(des).path+'/'+spath.name)
        
        dpath = moose.element(spath.path.replace(objS,objD))
        spoollist = moose.wildcardFind(spath.path+'/#[ISA=PoolBase]')
        dpoollist = moose.wildcardFind(dpath.path+'/#[ISA=PoolBase]')
        #check made, for a given Neutral or group if pool doesn't exist then copied
        # but some pool if enzyme cplx then there are holded untill that enzyme is copied in
        # `enzymeMerge` function        
        for spool in spoollist:
            if spool.name not in [dpool.name for dpool in dpoollist]:
                copied = copy_deleteUnlyingPoolObj(spool,dpath)
                if copied == False:
                    #hold it for later, this pool may be under enzyme, as cplx
                    poolNotcopiedyet.append(spool)
    
def copy_deleteUnlyingPoolObj(pool,path):
    # check if this pool is under compartement or under enzyme?(which is enzyme_cplx)
    # if enzyme_cplx then don't copy untill this perticular enzyme is copied
    # case: This enzyme_cplx might exist in modelA if enzyme exist then this
    # will automatically copie's the pool
    copied = False

    if pool.parent.className not in ["Enz","ZombieEnz","MMenz","ZombieMMenz"]:
         if path.className in ["Neutral","CubeMesh","CyclMesh"]:
            poolcopied = moose.copy(pool,path)
            copied = True
            # deleting function and enzyme which gets copied if exist under pool
            # This is done to ensure Dangling function / enzyme not copied.
            funclist = []
            for types in ['setConc','setN','increment']:
                funclist.extend(moose.element(poolcopied).neighbors[types])

            for fl in funclist:
                moose.delete(fl)
            enzlist = moose.element(poolcopied).neighbors['reac']
            for el in list(set(enzlist)):
                moose.delete(el.path)
    return copied

def updatePoolList(comptAdict):
    for key,value in list(comptAdict.items()):
        plist = moose.wildcardFind(value.path+'/##[ISA=PoolBase]')
        poolListina[key] = plist
    return poolListina

def enzymeMerge(comptD,comptS,key,poolListind):
    war_msg = ""
    RE_Duplicated, RE_Notcopiedyet, RE_Dangling = [], [], []
    comptDpath = moose.element(comptD[key]).path
    comptSpath = moose.element(comptS[key]).path
    objD = moose.element(comptDpath).parent.name
    objS = moose.element(comptSpath).parent.name
    #nzyListina => enzyListind

    enzyListind = moose.wildcardFind(comptDpath+'/##[ISA=EnzBase]')
    enzyListins = moose.wildcardFind(comptSpath+'/##[ISA=EnzBase]')
    for es in enzyListins:
        eSsubname, eSprdname = [],[]
        eSsubname = subprdList(es,"sub")
        eSprdname = subprdList(es,"prd")
        allexists, allexistp = False, False
        allclean = False

        poolinDlist = poolListind[findCompartment(es).name]
        for pD in poolinDlist:
            if es.parent.name == pD.name:
                edpath = es.parent.path.replace(objS,objD)

                if not moose.exists(edpath+'/'+es.name):
                    #This will take care
                    #  -- If same enzparent name but different enzyme name
                    #  -- or different parent/enzyme name
                    if eSsubname and eSprdname:
                        allexists = checkexist(eSsubname,objS,objD)
                        allexistp = checkexist(eSprdname,objS,objD)
                        if allexists and allexistp:
                            enzPool = moose.element(pD.path)
                            edpath = es.parent.path.replace(objS,objD)
                            enz = moose.element(moose.copy(es,moose.element(edpath)))
                            enzPool = enz.parent
                            if es.className in ["ZombieEnz","Enz"]:
                                moose.connect(moose.element(enz),"enz",enzPool,"reac")
                            if es.className in ["ZombieMMenz","MMenz"]:
                                moose.connect(enzPool,"nOut",enz,"enzDest")
                            connectObj(enz,eSsubname,"sub",comptD,war_msg)
                            connectObj(enz,eSprdname,"prd",comptD,war_msg)
                            allclean = True
                        else:
                            # didn't find sub or prd for this Enzyme
                            RE_Notcopiedyet.append(es)
                    else:
                        #   -- it is dagging reaction
                        RE_Dangling.append(es)
                else:
                    #Same Enzyme name
                    #   -- Same substrate and product including same volume then don't copy
                    #   -- different substrate/product or if sub/prd's volume is different then DUPLICATE the Enzyme
                    allclean = False
                    ed = moose.element(es.path.replace(objS,objD))
                    eDsubname = subprdList(ed,"sub")
                    eSsubname = subprdList(es,"sub")
                    hasSamenoofsublen,hasSameS,hasSamevols = same_len_name_vol(eDsubname,eSsubname)

                    eDprdname = subprdList(ed,"prd")
                    eSprdname = subprdList(es,"prd")
                    hasSamenoofprdlen,hasSameP,hasSamevolp = same_len_name_vol(eDprdname,eSprdname)
                    if not all((hasSamenoofsublen,hasSameS,hasSamevols,hasSamenoofprdlen,hasSameP,hasSamevolp)):
                        # May be different substrate or product or volume of Sub/prd may be different,
                        # Duplicating the enzyme
                        if eSsubname and eSprdname:
                            allexists,allexistp = False,False
                            allexists = checkexist(eSsubname,objS,objD)
                            allexistp = checkexist(eSprdname,objS,objD)
                            if allexists and allexistp:
                                es.name = es.name+"_duplicated"
                                if es.className in ["ZombieEnz","Enz"]:
                                    edpath = es.parent.path.replace(objS,objD)
                                    enz = moose.copy(es,moose.element(edpath))
                                    moose.connect(enz, 'enz', edpath, 'reac' )

                                if es.className in ["ZombieMMenz","MMenz"]:
                                    edpath = es.parent.path.replace(objS,objD)
                                    enz = moose.copy(es,moose.element(edpath))
                                    enzinfo = moose.Annotator(enz.path+'/info')
                                    moose.connect(moose.element(enz).parent,"nOut",moose.element(enz),"enzDest")
                                    #moose.connect(moose.element(enz),"enz",moose.element(enz).parent,"reac")

                                connectObj(enz,eSsubname,"sub",comptD,war_msg)
                                connectObj(enz,eSprdname,"prd",comptD,war_msg)
                                RE_Duplicated.append(enz)
                                allclean = True
                            else:
                                allclean = False
                    else:
                        allclean = True

                    if not allclean:
                        # didn't find sub or prd for this enzyme
                        #   --  it may be connected Enzyme cplx
                        if eSsubname and eSprdname:
                            RE_Notcopiedyet.append(es)
                        else:
                            RE_Dangling.append(es)

    return RE_Duplicated,RE_Notcopiedyet,RE_Dangling

def reacMerge(comptS,comptD,key,poolListina):
    RE_Duplicated, RE_Notcopiedyet, RE_Dangling = [], [], []
    war_msg = ""
    comptSpath = moose.element(comptS[key]).path
    comptDpath = moose.element(comptD[key]).path
    objS = moose.element(comptSpath).parent.name
    objD = moose.element(comptDpath).parent.name
    
    reacListins = moose.wildcardFind(comptSpath+'/##[ISA=ReacBase]')
    reacListind = moose.wildcardFind(comptDpath+'/##[ISA=ReacBase]')
    
    for rs in reacListins:
        rSsubname, rSprdname = [],[]
        rSsubname = subprdList(rs,"sub")
        rSprdname = subprdList(rs,"prd")
        allexists, allexistp = False, False
        allclean = False

        if rs.name not in [rd.name for rd in reacListind]:
            # reaction name not found then copy
            # And assuming that pools are copied earlier EXPECT POOL CPLX
            # To be assured the it takes correct compartment name incase reaction sub's
            # belongs to different compt
            key = findCompartment(rs).name
            if rSsubname and rSprdname:
                allexists =  checkexist(rSsubname,objS,objD)
                allexistp =  checkexist(rSprdname,objS,objD)
                if allexists and allexistp:
                    rdpath = rs.parent.path.replace(objS,objD)
                    reac = moose.copy(rs,moose.element(rdpath))
                    connectObj(reac,rSsubname,"sub",comptD,war_msg)
                    connectObj(reac,rSprdname,"prd",comptD,war_msg)
                    allclean = True
                else:
                    # didn't find sub or prd for this reaction
                    #   --  it may be connected Enzyme cplx
                    RE_Notcopiedyet.append(rs)
            else:
                #   -- it is dagging reaction
                RE_Dangling.append(rs)
                #print ("This reaction \""+rb.path+"\" has no substrate/product Dangling reaction are not copied")
                #war_msg = war_msg+"\nThis reaction \""+rb.path+"\" has no substrate/product Dangling reaction are not copied"

        else:
            #Same reaction name
            #   -- Same substrate and product including same volume then don't copy
            #   -- different substrate/product or if sub/prd's volume is different then DUPLICATE the reaction
            
            allclean = False
            for rd in reacListind:
                if rs.name == rd.name:
                    rSsubname = subprdList(rs,"sub")
                    rDsubname = subprdList(rd,"sub")
                    hasSamenoofsublen,hasSameS,hasSamevols = same_len_name_vol(rSsubname,rDsubname)

                    rSprdname = subprdList(rs,"prd")
                    rDprdname = subprdList(rd,"prd")
                    hasSamenoofprdlen,hasSameP,hasSamevolp = same_len_name_vol(rSprdname,rDprdname)
                    if not all((hasSamenoofsublen,hasSameS,hasSamevols,hasSamenoofprdlen,hasSameP,hasSamevolp)):
                        # May be different substrate or product or volume of Sub/prd may be different,
                        # Duplicating the reaction
                        if rSsubname and rSprdname:
                            allexists,allexistp = False,False
                            allexists = checkexist(rSsubname,objS,objD)
                            allexistp = checkexist(rSprdname,objS,objD)
                            if allexists and allexistp:
                                rs.name = rs.name+"_duplicated"
                                #reac = moose.Reac(comptA[key].path+'/'+rb.name+"_duplicated")
                                rdpath = rs.parent.path.replace(objS,objD)
                                reac = moose.copy(rs,moose.element(rdpath))
                                connectObj(reac,rSsubname,"sub",comptD,war_msg)
                                connectObj(reac,rSprdname,"prd",comptD,war_msg)
                                RE_Duplicated.append(reac)
                                allclean = True
                            else:
                                allclean = False
                    else:
                        allclean = True

                    if not allclean:
                        # didn't find sub or prd for this reaction
                        #   --  it may be connected Enzyme cplx
                        if rSsubname and rSprdname:
                            RE_Notcopiedyet.append(rs)
                        else:
                            RE_Dangling.append(rs)
    
    return RE_Duplicated,RE_Notcopiedyet,RE_Dangling

def subprdList(reac,subprd):
    rtype = moose.element(reac).neighbors[subprd]
    rname = []
    for rs in rtype:
        rname.append(moose.element(rs))
    return rname

def same_len_name_vol(rA,rB):
    uaS = set(rA)
    ubS = set(rB)
    aS = set([uas.name for uas in uaS])
    bS = set([ubs.name for ubs in ubS])

    hassameLen = False
    hassameSP  = False
    hassamevol = False
    hassamevollist = []
    if (len(rA) == len(rB) ):
        hassameLen = True
        if  not (len (aS.union(bS) - aS.intersection(bS))):
            hassameSP = True
            if rB and rA:
                rAdict = dict( [ (i.name,i) for i in (rA) ] )
                rBdict = dict( [ (i.name,i) for i in (rB) ] )

                for key,bpath in rBdict.items():
                    apath = rAdict[key]
                    comptA = moose.element(findCompartment(apath))
                    comptB = moose.element(findCompartment(bpath))
                    if not abs(comptA.volume -comptB.volume):
                            hassamevollist.append(True)
                    else:
                            hassamevollist.append(False)

    if len(set(hassamevollist))==1:
        for x in set(hassamevollist):
            hassamevol = x

    return ( hassameLen,hassameSP,hassamevol)

def connectObj(reac,spList,spType,comptA,war_msg):
    #It should not come here unless the sub/prd is connected to enzyme cplx pool
    allclean = False
    for rsp in spList:
        for akey in list(poolListina[findCompartment(rsp).name]):
            if rsp.name == akey.name:
                if moose.exists(akey.path):
                    moose.connect(moose.element(reac), spType, moose.element(akey), 'reac', 'OneToOne')
                    allclean = True
                else:
                    #It should not come here unless the sub/prd is connected to enzyme cplx pool
                    allclean = False
    return allclean

def checkexist(spList,objB,objA):
    allexistL = []
    allexist = False
    for rsp in spList:
        found = False
        rspPath = rsp.path.replace(objB,objA)
        if moose.exists(rspPath):
            found = True
        allexistL.append(found)

    if len(set(allexistL))==1:
        for x in set(allexistL):
            allexist = x

    return allexist

if __name__ == "__main__":
    try:
        sys.argv[1]
    except IndexError:
        print("Source filename or path not given")
        exit(0)
    else:
        src = sys.argv[1]
        if not os.path.exists(src):
            print("Filename or path does not exist %s." %(src))
        else:
            try:
                sys.argv[2]
            except IndexError:
                print("Destination filename or path not given")
                exit(0)
            else:
                des = sys.argv[2]
                if not os.path.exists(src):
                    print("Filename or path does not exist %s." %(des))
                    exit(0)
                else:
                    print ("src and des %s, %s." %(src, des))
                    mergered = mergeChemModel(src,des)

                '''
                try:
                    sys.argv[3]
                except IndexError:
                    print ("Merge to save not specified")
                    mergeto = "merge"
                else:
                    mergeto = sys.argv[3]
                    if str(mergeto).rfind('.') != -1:
                        mergeto = mergeto[:str(mergeto).rfind('.')]
                    if str(mergeto).rfind('/'):
                        mergeto = mergeto+'merge'

                    mergered = mergeChemModel(src,des,mergeto)
                '''



# -*- coding: utf-8 -*-
#########################################################################
## This program is part of 'MOOSE', the
## Messaging Object Oriented Simulation Environment.
##           Copyright (C) 2018 Upinder S. Bhalla. and NCBS
## It is made available under the terms of the
## GNU General Public License version 3 or later.
## See the file COPYING.LIB for the full notice.
##
## This file makes u + t <=====> s where u is in a different compt from t,s
#########################################################################


import math
import numpy as np
import moose
import moose.fixXreacs as fixXreacs

def makeModel():
    # create container for model
    num = 1 # number of compartments
    model = moose.Neutral( '/model' )
    compartment = moose.CylMesh( '/model/compartment' )
    compartment.x1 = 1.0e-6 # Set it to a 1 micron single-voxel cylinder

    # create molecules and reactions
    u = moose.Pool( '/model/compartment/u' )
    #####################################################################
    # Put in endo compartment. Add molecule s
    endo = moose.EndoMesh( '/model/endo' )
    endo.isMembraneBound = True
    endo.surround = compartment
    rXfer = moose.Reac( '/model/endo/rXfer' )
    et = moose.Pool( '/model/endo/t' )
    es = moose.Pool( '/model/endo/s' )
    #####################################################################
    moose.connect( rXfer, 'sub', u, 'reac' )
    moose.connect( rXfer, 'sub', et, 'reac' )
    moose.connect( rXfer, 'prd', es, 'reac' )
    u.concInit = 1.0
    et.concInit = 1.0

    #####################################################################
    # [u0] = 1 in compt, [t0] = 1 in endo, [s0] = 0
    # [u] + [s]/8 = [u0]  ; [t] + [s] = [t0]; nu + ns = nu0, nt + ns = nt0
    # At equil, numKf*nu*nt = numKb*ns
    # Express all # in terms of ns.
    # nu = nu0-ns; nt = nt0-ns      Also, nu0 = 8*nt0
    # So  numKf*(nu0-ns)*(nt0-ns) = numKb*ns
    # Target level is nt = ns = nt0/2
    # So numKf*( 8nt0 - nt0/2)*nt0/2 = numKb*nt0/2
    # So 7.5*nt0 = numKb/numKf
    rXfer.numKb = 0.1
    rXfer.numKf = 0.1/(7.5 * et.nInit)
    #print( "Rates = {}, {}".format( rXfer.Kf, rXfer.Kb ))
    #####################################################################
    fixXreacs.fixXreacs( '/model' )
    #fixXreacs.restoreXreacs( '/model' )
    #fixXreacs.fixXreacs( '/model' )
    #####################################################################

    # Make solvers
    ksolve = moose.Ksolve( '/model/compartment/ksolve' )
    dsolve = moose.Dsolve( '/model/compartment/dsolve' )
    eksolve = moose.Ksolve( '/model/endo/ksolve' )
    edsolve = moose.Dsolve( '/model/endo/dsolve' )

    stoich = moose.Stoich( '/model/compartment/stoich' )
    stoich.compartment = compartment
    stoich.ksolve = ksolve
    stoich.dsolve = dsolve
    stoich.path = "/model/compartment/##"
    assert( dsolve.numPools == 1 )

    estoich = moose.Stoich( '/model/endo/stoich' )
    estoich.compartment = endo
    estoich.ksolve = eksolve
    estoich.dsolve = edsolve
    estoich.path = "/model/endo/##"
    assert( edsolve.numPools == 3 )

    edsolve.buildMeshJunctions( dsolve )

    plot1 = moose.Table2( '/model/plot1' )
    plot2 = moose.Table2( '/model/plot2' )
    plot3 = moose.Table2( '/model/plot3' )
    moose.connect( plot1, 'requestOut', u, 'getN' )
    moose.connect( plot2, 'requestOut', et, 'getN' )
    moose.connect( plot3, 'requestOut', es, 'getN' )
    plot4 = moose.Table2( '/model/plot4' )
    plot5 = moose.Table2( '/model/plot5' )
    plot6 = moose.Table2( '/model/plot6' )
    moose.connect( plot4, 'requestOut', u, 'getConc' )
    moose.connect( plot5, 'requestOut', et, 'getConc' )
    moose.connect( plot6, 'requestOut', es, 'getConc' )

def almostEq( a, b ):
    #print a, b, (a-b)/(a+b)
    return abs(a-b)/(a+b) < 5e-5

def test_xreacs7():
    runtime = 100
    makeModel()
    moose.reinit()
    moose.start( runtime )

    assert( almostEq( moose.element( 'model/endo/t' ).conc,
        moose.element( '/model/endo/s' ).conc ) )
    moose.delete( '/model' )

# Run the 'main' if this script is executed standalone.
if __name__ == '__main__':
    test_xreacs7()



# -*- coding: utf-8 -*-
"""
Utility to create mapping from NeuroML2 unit strings (as specified
in NeuroMLCoreDimensions.xml) to SI

Author: 
    Subhasis Ray (Please check commit history for more details).
"""

from __future__ import print_function, division
import os
import re
from xml.etree import ElementTree

# The units and dimensions used in NeuroML2 are defined in this file:
# https://github.com/NeuroML/NeuroML2/blob/master/NeuroML2CoreTypes/NeuroMLCoreDimensions.xml
unitsdoc = ElementTree.parse(os.path.join(os.path.dirname(__file__),'schema', 'NeuroMLCoreDimensions.xml'))
dims = dict([(el.attrib['name'], el) for el in unitsdoc.getroot() if el.tag == '{http://www.neuroml.org/lems/0.7.4}Dimension'])
units = dict([(el.attrib['symbol'], el) for el in unitsdoc.getroot() if el.tag == '{http://www.neuroml.org/lems/0.7.4}Unit'])

# optional +/- followed by 0 or more digits followed by optional
# decimal point followed by optional digits followed by optional
# (exponent symbol followed by optional +/- followed by one or more
# digits.
magnitude_regex = re.compile(r'^[\+-]?\d*\.?\d*([eE][\+-]?\d+)?')

def SI(expression):
    try:
        return float(expression)
    except:
        pass
    expression=expression.replace(" ", "")
    match = magnitude_regex.match(expression)
    magnitude = float(match.group(0))
    unitstr = re.split(magnitude_regex, expression)[-1]
    
    #print("Converting %s: mag: %s, unitstr: %s"%(expression, magnitude, unitstr))
    try:
        unit = units[unitstr]
        #print('Unit: %s'%unit.attrib)
    except KeyError as ke:
        print("Error in converting %s: %s, using %s"%(expression, ke, magnitude))
        return magnitude

    power = int(unit.attrib['power']) if 'power' in unit.attrib else 0
    offset = float(unit.attrib['offset']) if 'offset' in unit.attrib else 0
    scale = float(unit.attrib['scale']) if 'scale' in unit.attrib else 1
    
    si = (magnitude + offset) * scale * 10**power 

    return si


if __name__ == "__main__":
    examples = ['-70mV','5 V','1',2,3.3, '330mM','15K','0 degC','-300degC','1min']
    for e in examples:
        print(SI(e))
#
# units.py ends here



# Modified from function.py ---

import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

import moose

simtime = 1.0

def test_example():
    moose.Neutral('/model')
    function = moose.Function('/model/function')
    function.c['c0'] = 1.0
    function.c['c1'] = 2.0
    #function.x.num = 1
    function.expr = 'c0 * exp(c1*x0) * cos(y0) + sin(t)'
    # mode 0 - evaluate function value, derivative and rate
    # mode 1 - just evaluate function value,
    # mode 2 - evaluate derivative,
    # mode 3 - evaluate rate
    function.mode = 0
    function.independent = 'y0'
    nsteps = 1000
    xarr = np.linspace(0.0, 1.0, nsteps)
    # Stimulus tables allow you to store sequences of numbers which
    # are delivered via the 'output' message at each time step. This
    # is a placeholder and in real scenario you will be using any
    # sourceFinfo that sends out a double value.
    input_x = moose.StimulusTable('/xtab')
    input_x.vector = xarr
    input_x.startTime = 0.0
    input_x.stepPosition = xarr[0]
    input_x.stopTime = simtime
    moose.connect(input_x, 'output', function.x[0], 'input')

    yarr = np.linspace(-np.pi, np.pi, nsteps)
    input_y = moose.StimulusTable('/ytab')
    input_y.vector = yarr
    input_y.startTime = 0.0
    input_y.stepPosition = yarr[0]
    input_y.stopTime = simtime
    moose.connect(function, 'requestOut', input_y, 'getOutputValue')

    # data recording
    result = moose.Table('/ztab')
    moose.connect(result, 'requestOut', function, 'getValue')
    derivative = moose.Table('/zprime')
    moose.connect(derivative, 'requestOut', function, 'getDerivative')
    rate = moose.Table('/dz_by_dt')
    moose.connect(rate, 'requestOut', function, 'getRate')
    x_rec = moose.Table('/xrec')
    moose.connect(x_rec, 'requestOut', input_x, 'getOutputValue')
    y_rec = moose.Table('/yrec')
    moose.connect(y_rec, 'requestOut', input_y, 'getOutputValue')

    dt =  simtime/nsteps
    for ii in range(32):
        moose.setClock(ii, dt)
    moose.reinit()
    moose.start(simtime)

    # Uncomment the following lines and the import matplotlib.pyplot as plt on top
    # of this file to display the plot.
    plt.subplot(3,1,1)
    plt.plot(x_rec.vector, result.vector, 'r-', label='z = {}'.format(function.expr))
    z = function.c['c0'] * np.exp(function.c['c1'] * xarr) * np.cos(yarr) + np.sin(np.arange(len(xarr)) * dt)
    plt.plot(xarr, z, 'b--', label='numpy computed')
    plt.xlabel('x')
    plt.ylabel('z')
    plt.legend()

    plt.subplot(3,1,2)
    plt.plot(y_rec.vector, derivative.vector, 'r-', label='dz/dy0')
    # derivatives computed by putting x values in the analytical formula
    dzdy = function.c['c0'] * np.exp(function.c['c1'] * xarr) * (- np.sin(yarr))
    plt.plot(yarr, dzdy, 'b--', label='numpy computed')
    plt.xlabel('y')
    plt.ylabel('dz/dy')
    plt.legend()

    plt.subplot(3,1,3)
    # *** BEWARE *** The first two entries are spurious. Entry 0 is
    # *** from reinit sending out the defaults. Entry 2 is because
    # *** there is no lastValue for computing real forward difference.
    plt.plot(np.arange(2, len(rate.vector), 1) * dt, rate.vector[2:], 'r-', label='dz/dt')
    dzdt = np.diff(z)/dt
    plt.plot(np.arange(0, len(dzdt), 1.0) * dt, dzdt, 'b--', label='numpy computed')
    plt.xlabel('t')
    plt.ylabel('dz/dt')
    plt.legend()
    plt.tight_layout()
    plt.savefig(__file__+'.png')

if __name__ == '__main__':
    test_example()



# -*- coding: utf-8 -*-
from __future__ import print_function
import moose

moose.Neutral('/model')
compt = moose.CubeMesh('/model/Compart')
compt.volume = 1.6667e-21

s = moose.Pool('/model/Compart/S')
s.concInit = 0.3
p = moose.Pool('/model/Compart/P')
r = moose.Reac('/model/Compart/Reac')
moose.connect(r, 'sub', s, 'reac')

moose.connect(r, 'prd', p, 'reac')

bf = moose.BufPool('/model/Compart/BufPool')
f = moose.Function('/model/Compart/BufPool/function')
moose.connect( f, 'valueOut', bf ,'setN' )

numVariables = f.numVars
f.numVars+=1
expr = ""
expr = (f.expr+'+'+'x'+str(numVariables))
expr = expr.lstrip("0 +")
expr = expr.replace(" ","")
f.expr = expr
moose.connect( s, 'nOut', f.x[numVariables], 'input' )

bf1 = moose.BufPool('/model/Compart/BufPool1')
f1 = moose.Function('/model/Compart/BufPool1/func')
moose.connect( f1, 'valueOut', bf1 ,'setN' )
numVariables = f1.numVars
expr = ""
expr = (f1.expr+'+'+'x'+str(numVariables))
expr = expr.lstrip("0 +")
expr = expr.replace(" ","")
f1.expr = expr
moose.connect( s, 'nOut', f1.x[numVariables], 'input' )

compts = moose.wildcardFind('/model/##[ISA=ChemCompt]')

print(" f name= ",f.name, f.tick)
print(" f1 name= ",f1.name, f1.tick)
for compt in compts:
    ksolve = moose.Ksolve( compt.path+'/ksolve' )
    stoich = moose.Stoich( compt.path+'/stoich' )
    stoich.compartment = compt
    stoich.ksolve = ksolve
    stoich.path = compt.path+"/##"

print("After solver is set")
print(" f name= ",f.name,f.tick)
print(" f1 name= ", f1.name, f1.tick)



# -*- coding: utf-8 -*-
## Description: class ChannelML for loading ChannelML from file or xml element into MOOSE
## Version 1.0 by Aditya Gilra, NCBS, Bangalore, India, 2011 for serial MOOSE
## Version 1.5 by Niraj Dudani, NCBS, Bangalore, India, 2012, ported to parallel MOOSE
## Version 1.6 by Aditya Gilra, NCBS, Bangalore, India, 2012, minor changes for parallel MOOSE
## Version 1.7 by Aditya Gilra, NCBS, Bangalore, India, 2013, further support for NeuroML 1.8.1
## Version 1.8 by Aditya Gilra, NCBS, Bangalore, India, 2013, changes for new IntFire and SynHandler classes

"""
NeuroML.py is the preferred interface. Use this only if NeuroML L1,L2,L3 files are misnamed/scattered.
Instantiate ChannelML class, and thence use method:
readChannelMLFromFile(...) to load a standalone ChannelML file (synapse/channel), OR
readChannelML(...) / readSynapseML to load from an xml.etree xml element (could be part of a larger NeuroML file).
"""

from __future__ import print_function, division
from xml.etree import cElementTree as ET
import string
import os
import sys
import math
import moose
from moose.neuroml import utils
import moose.utils as mu
import moose.print_utils as pu

class ChannelML():

    def __init__(self,nml_params):

        self.cml='http://morphml.org/channelml/schema'
        self.temperature = nml_params['temperature']

    def readChannelMLFromFile(self,filename,params={}):
        """ specify params as a dict.
            Currently none implemented.
            For setting channel temperature,
             pass in via nml_params during instance creation (see __init__())
             or set <instance>.temperature."""
        tree = ET.parse(filename)
        channelml_element = tree.getroot()
        for channel in channelml_element.findall('.//{'+self.cml+'}channel_type'):
            ## ideally I should read in extra params from within the channel_type element
            ## and put those in also. Global params should override local ones.
            self.readChannelML(channel,params,channelml_element.attrib['units'])
        for synapse in channelml_element.findall('.//{'+self.cml+'}synapse_type'):
            self.readSynapseML(synapse,channelml_element.attrib['units'])
        for ionConc in channelml_element.findall('.//{'+self.cml+'}ion_concentration'):
            self.readIonConcML(ionConc,channelml_element.attrib['units'])

    def readSynapseML(self,synapseElement,units="SI units"):
        if 'Physiological Units' in units: # see pg 219 (sec 13.2) of Book of Genesis
            Vfactor = 1e-3 # V from mV
            Tfactor = 1e-3 # s from ms
            Gfactor = 1e-3 # S from mS
        elif 'SI Units' in units:
            Vfactor = 1.0
            Tfactor = 1.0
            Gfactor = 1.0
        else:
            pu.fatal("Wrong units %s exiting ..." % units)
            sys.exit(1)
        if not moose.exists('/library'):
            moose.Neutral('/library') 
        synname = synapseElement.attrib['name']
        if utils.neuroml_debug:
           pu.info("Loading synapse : %s into /library" % synname)
        moosesynapse = moose.SynChan('/library/'+synname)
        doub_exp_syn = synapseElement.find('./{'+self.cml+'}doub_exp_syn')
        moosesynapse.Ek = float(doub_exp_syn.attrib['reversal_potential'])*Vfactor
        moosesynapse.Gbar = float(doub_exp_syn.attrib['max_conductance'])*Gfactor
        moosesynapse.tau1 = float(doub_exp_syn.attrib['rise_time'])*Tfactor # seconds
        moosesynapse.tau2 = float(doub_exp_syn.attrib['decay_time'])*Tfactor # seconds
        ## The delay and weight can be set only after connecting a spike event generator.
        ## delay and weight are arrays: multiple event messages can be connected to a single synapse

        ## graded synapses are not supported by neuroml, so set to False here,
        ## see my Demo/neuroml/lobster_pyloric/STG_net.py for how to still have graded synapses
        moosesynapse_graded = moose.Mstring(moosesynapse.path+'/graded')
        moosesynapse_graded.value = 'False'
        moosesynapse_mgblock = moose.Mstring(moosesynapse.path+'/mgblockStr')
        moosesynapse_mgblock.value = 'False'
        ## check if STDP synapse is present or not
        stdp_syn = synapseElement.find('./{'+self.cml+'}stdp_syn')
        if stdp_syn is None:
            moosesynhandler = moose.SimpleSynHandler('/library/'+synname+'/handler')
        else:
            moosesynhandler = moose.STDPSynHandler('/library/'+synname+'/handler')
            moosesynhandler.aPlus0 = float(stdp_syn.attrib['del_weight_ltp'])
            moosesynhandler.aMinus0 = float(stdp_syn.attrib['del_weight_ltd'])
            moosesynhandler.tauPlus = float(stdp_syn.attrib['tau_ltp'])
            moosesynhandler.tauMinus = float(stdp_syn.attrib['tau_ltd'])
            moosesynhandler.weightMax = float(stdp_syn.attrib['max_syn_weight'])
            moosesynhandler.weightMin = 0.0
        ## connect the SimpleSynHandler or the STDPSynHandler to the SynChan (double exp)
        moose.connect( moosesynhandler, 'activationOut', moosesynapse, 'activation' )

    def readChannelML(self,channelElement,params={},units="SI units"):
        ## I first calculate all functions assuming a consistent system of units.
        ## While filling in the A and B tables, I just convert to SI.
        ## Also convert gmax and Erev.
        if 'Physiological Units' in units: # see pg 219 (sec 13.2) of Book of Genesis
            Vfactor = 1e-3 # V from mV
            Tfactor = 1e-3 # s from ms
            Gfactor = 1e1 # S/m^2 from mS/cm^2
            concfactor = 1e6 # Mol = mol/m^-3 from mol/cm^-3
        elif 'SI Units' in units:
            Vfactor = 1.0
            Tfactor = 1.0
            Gfactor = 1.0
            concfactor = 1.0
        else:
            raise RuntimeError("Wrong units %s. Existing" % units)

        if not moose.exists('/library'):
            moose.Neutral('/library')

        channel_name = channelElement.attrib['name']
        if utils.neuroml_debug:
           pu.info("Loading channel %s into /library" % channel_name)

        IVrelation = channelElement.find('./{'+self.cml+'}current_voltage_relation')
        intfire = IVrelation.find('./{'+self.cml+'}integrate_and_fire')

        if intfire is not None:
            ## Below params need to be set while making an LIF compartment
            moosechannel = moose.Neutral('/library/'+channel_name)
            moosechannelval = moose.Mstring(moosechannel.path+'/vReset')
            moosechannelval.value = str(float(intfire.attrib['v_reset'])*Vfactor)
            moosechannelval = moose.Mstring(moosechannel.path+'/thresh')
            moosechannelval.value = str(float(intfire.attrib['threshold'])*Vfactor)
            moosechannelval = moose.Mstring(moosechannel.path+'/refracT')
            moosechannelval.value = str(float(intfire.attrib['t_refrac'])*Tfactor)
            ## refracG is currently not supported by moose.LIF
            ## Confirm if g_refrac is a conductance density or not?
            ## assuming g_refrac is a conductance density below
            moosechannelval = moose.Mstring(moosechannel.path+'/refracG')
            moosechannelval.value = str(float(intfire.attrib['g_refrac'])*Gfactor)
            ## create an Mstring saying this is an integrate_and_fire mechanism
            moosechannelval = moose.Mstring(moosechannel.path+'/integrate_and_fire')
            moosechannelval.value = 'True'
            return

        concdep = IVrelation.find('./{'+self.cml+'}conc_dependence')
        if concdep is None:
            moosechannel = moose.HHChannel('/library/'+channel_name)
        else:
            moosechannel = moose.HHChannel2D('/library/'+channel_name)

        if IVrelation.attrib['cond_law']=="ohmic":
            moosechannel.Gbar = float(IVrelation.attrib['default_gmax']) * Gfactor
            moosechannel.Ek = float(IVrelation.attrib['default_erev']) * Vfactor
            moosechannelIon = moose.Mstring(moosechannel.path+'/ion')
            moosechannelIon.value = IVrelation.attrib['ion']
            if concdep is not None:
                moosechannelIonDependency = moose.Mstring(moosechannel.path+'/ionDependency')
                moosechannelIonDependency.value = concdep.attrib['ion']

        nernstnote = IVrelation.find('./{'+utils.meta_ns+'}notes')
        if nernstnote is not None:
            ## the text in nernstnote is "Nernst,Cout=<float>,z=<int>"
            nernst_params = nernstnote.text.split(',')
            if nernst_params[0] == 'Nernst':
                nernstMstring = moose.Mstring(moosechannel.path+'/nernst_str')
                nernstMstring.value = str( float(nernst_params[1].split('=')[1]) * concfactor ) + \
                                        ',' + str( int(nernst_params[2].split('=')[1]) )

        gates = IVrelation.findall('./{'+self.cml+'}gate')
        if len(gates)>3:
            pu.fatal("Sorry! Maximum x, y, and z (three) gates are possible in MOOSE/Genesis")
            sys.exit()
        gate_full_name = [ 'gateX', 'gateY', 'gateZ' ] # These are the names that MOOSE uses to create gates.
        ## if impl_prefs tag is present change VMIN, VMAX and NDIVS
        impl_prefs = channelElement.find('./{'+self.cml+'}impl_prefs')
        if impl_prefs is not None:
            table_settings = impl_prefs.find('./{'+self.cml+'}table_settings')
            ## some problem here... disable
            VMIN_here = float(table_settings.attrib['min_v'])
            VMAX_here = float(table_settings.attrib['max_v'])
            NDIVS_here = int(table_settings.attrib['table_divisions'])
            dv_here = (VMAX_here - VMIN_here) / NDIVS_here
        else:
            ## default VMIN, VMAX and dv are in SI
            ## convert them to current calculation units used by channel definition
            ## while loading into tables, convert them back to SI
            VMIN_here = utils.VMIN/Vfactor
            VMAX_here = utils.VMAX/Vfactor
            NDIVS_here = utils.NDIVS
            dv_here = utils.dv/Vfactor
        offset = IVrelation.find('./{'+self.cml+'}offset')
        if offset is None: vNegOffset = 0.0
        else: vNegOffset = float(offset.attrib['value'])
        self.parameters = []
        for parameter in channelElement.findall('.//{'+self.cml+'}parameter'):
            self.parameters.append( (parameter.attrib['name'],float(parameter.attrib['value'])) )

        for num,gate in enumerate(gates):
            # if no q10settings tag, the q10factor remains 1.0
            # if present but no gate attribute, then set q10factor
            # if there is a gate attribute, then set it only if gate attrib matches gate name
            self.q10factor = 1.0
            self.gate_name = gate.attrib['name']
            for q10settings in IVrelation.findall('./{'+self.cml+'}q10_settings'):
                ## self.temperature from neuro.utils
                if 'gate' in q10settings.attrib:
                    if q10settings.attrib['gate'] == self.gate_name:
                        self.setQ10(q10settings)
                        break
                else:
                    self.setQ10(q10settings)

            ############### HHChannel2D crashing on setting Xpower!
            #### temperamental! If you print something before, it gives cannot creategate from copied channel, else crashes
            ## Setting power first. This is necessary because it also
            ## initializes the gate's internal data structures as a side
            ## effect. Alternatively, gates can be initialized explicitly
            ## by calling HHChannel.createGate().
            gate_power = float( gate.get( 'instances' ) )
            if num == 0:
                moosechannel.Xpower = gate_power
                if concdep is not None: moosechannel.Xindex = "VOLT_C1_INDEX"
            elif num == 1:
                moosechannel.Ypower = gate_power
                if concdep is not None: moosechannel.Yindex = "VOLT_C1_INDEX"
            elif num == 2:
                moosechannel.Zpower = gate_power
                if concdep is not None: moosechannel.Zindex = "VOLT_C1_INDEX"

            ## Getting handle to gate using the gate's path.
            gate_path = moosechannel.path + '/' + gate_full_name[ num ]
            if concdep is None:
                if not moose.exists(gate_path):
                    moosegate = moose.HHGate(gate_path)
                else:
                    moosegate = moose.element(gate_path)
                ## set SI values inside MOOSE
                moosegate.min = VMIN_here*Vfactor
                moosegate.max = VMAX_here*Vfactor
                moosegate.divs = NDIVS_here
                ## V.IMP to get smooth curves, else even with 3000 divisions
                ## there are sudden transitions.
                moosegate.useInterpolation = True
            else:
                moosegate = moose.HHGate2D( gate_path )

            ##### If alpha and beta functions exist, make them here
            for transition in gate.findall('./{'+self.cml+'}transition'):
                ## make python functions with names of transitions...
                fn_name = transition.attrib['name']
                ## I assume that transitions if present are called alpha and beta
                ## for forward and backward transitions...
                if fn_name in ['alpha','beta']:
                    self.make_cml_function(transition, fn_name, concdep)
                else:
                    pu.fatal("Unsupported transition %s" % fn_name)
                    sys.exit()

            time_course = gate.find('./{'+self.cml+'}time_course')
            ## tau is divided by self.q10factor in make_function()
            ## thus, it gets divided irrespective of <time_course> tag present or not.
            if time_course is not None:
                self.make_cml_function(time_course, 'tau', concdep)
            steady_state = gate.find('./{'+self.cml+'}steady_state')
            if steady_state is not None:
                self.make_cml_function(steady_state, 'inf', concdep)

            if concdep is None: ca_name = ''                        # no Ca dependence
            else: ca_name = ','+concdep.attrib['variable_name']     # Ca dependence

            ## Create tau() and inf() if not present, from alpha() and beta()
            for fn_element,fn_name,fn_expr in [(time_course,'tau',"1/(alpha+beta)"),\
                                                (steady_state,'inf',"alpha/(alpha+beta)")]:
                ## put in args for alpha and beta, could be v and Ca dep.
                expr_string = fn_expr.replace('alpha', 'self.alpha(v'+ca_name+')')
                expr_string = expr_string.replace('beta', 'self.beta(v'+ca_name+')')
                ## if time_course/steady_state are not present,
                ## then alpha annd beta transition elements should be present, and fns created.
                if fn_element is None:
                        self.make_function(fn_name,'generic',expr_string=expr_string,concdep=concdep)

            ## non Ca dependent channel
            if concdep is None:
                ## while calculating, use the units used in xml defn,
                ## while filling in table, I convert to SI units.
                v0 = VMIN_here - vNegOffset
                n_entries = NDIVS_here+1
                tableA = [ 0.0 ] * n_entries
                tableB = [ 0.0 ] * n_entries
                for i in range(n_entries):
                    v = v0 + i * dv_here

                    inf = self.inf(v)
                    tau = self.tau(v)
                    ## convert to SI before writing to table
                    ## qfactor is already in inf and tau
                    tableA[i] = inf/tau / Tfactor
                    tableB[i] = 1.0/tau / Tfactor

                moosegate.tableA = tableA
                moosegate.tableB = tableB

            ## Ca dependent channel
            else:
                ## UNITS: while calculating, use the units used in xml defn,
                ##        while filling in table, I convert to SI units.
                ##        Note here Ca units do not enter, but
                ##         units of CaMIN, CaMAX and ca_conc in fn expr should match.
                v = VMIN_here - vNegOffset
                CaMIN = float(concdep.attrib['min_conc'])
                CaMAX = float(concdep.attrib['max_conc'])
                CaNDIVS = 100
                dCa = (CaMAX-CaMIN)/CaNDIVS
                ## CAREFUL!: tableA = [[0.0]*(CaNDIVS+1)]*(NDIVS_here+1) will not work!
                ## * does a shallow copy, same list will get repeated 200 times!
                ## Thus setting tableA[35][1] = 5.0 will set all rows, 1st col to 5.0!!!!
                tableA = [[0.0]*(CaNDIVS+1) for i in range(NDIVS_here+1)]
                tableB = [[0.0]*(CaNDIVS+1) for i in range(NDIVS_here+1)]
                for i in range(NDIVS_here+1):
                    Ca = CaMIN
                    for j in range(CaNDIVS+1):
                        inf = self.inf(v,Ca)
                        tau = self.tau(v,Ca)
                        ## convert to SI (Tfactor) before writing to table
                        ## qfactor is already in inf and tau
                        tableA[i][j] = inf/tau / Tfactor
                        tableB[i][j] = 1.0/tau / Tfactor
                        Ca += dCa
                    v += dv_here

                ## Presently HHGate2D doesn't allow the setting of tables as 2D vectors directly
                moosegate.tableA = tableA
                moosegate.tableB = tableB

                ## set SI values inside MOOSE
                moosegate.xminA = VMIN_here*Vfactor
                moosegate.xmaxA = VMAX_here*Vfactor
                moosegate.xdivsA = NDIVS_here
                #moosegate.dxA = dv_here*Vfactor
                moosegate.yminA = CaMIN*concfactor
                moosegate.ymaxA = CaMAX*concfactor
                moosegate.ydivsA = CaNDIVS
                #moosegate.dyB = dCa*concfactor

                ## set SI values inside MOOSE
                moosegate.xminB = VMIN_here*Vfactor
                moosegate.xmaxB = VMAX_here*Vfactor
                moosegate.xdivsB = NDIVS_here
                #moosegate.dxB = dv_here*Vfactor
                moosegate.yminB = CaMIN*concfactor
                moosegate.ymaxB = CaMAX*concfactor
                moosegate.ydivsB = CaNDIVS
                #moosegate.dyB = dCa*concfactor

    def setQ10(self,q10settings):
        if 'q10_factor' in q10settings.attrib:
            self.q10factor = float(q10settings.attrib['q10_factor'])\
                **((self.temperature-float(q10settings.attrib['experimental_temp']))/10.0)
        elif 'fixed_q10' in q10settings.attrib:
            self.q10factor = float(q10settings.attrib['fixed_q10'])


    def readIonConcML(self, ionConcElement, units="SI units"):
        if units == 'Physiological Units': # see pg 219 (sec 13.2) of Book of Genesis
            Vfactor = 1e-3 # V from mV
            Tfactor = 1e-3 # s from ms
            Gfactor = 1e1 # S/m^2 from mS/cm^2
            concfactor = 1e6 # mol/m^3 from mol/cm^3
            Lfactor = 1e-2 # m from cm
            Ifactor = 1e-6 # A from microA
        else:
            Vfactor = 1.0
            Tfactor = 1.0
            Gfactor = 1.0
            concfactor = 1.0
            Lfactor = 1.0
            Ifactor = 1.0
        moose.Neutral('/library') # creates /library in MOOSE tree; elif present, wraps
        ionSpecies = ionConcElement.find('./{'+self.cml+'}ion_species')
        if ionSpecies is not None:
            if not 'ca' in ionSpecies.attrib['name']:
                pu.fatal("Sorry, I cannot handle non-Ca-ion pools. Exiting ...")
                sys.exit(1)
        capoolName = ionConcElement.attrib['name']

        pu.info("Loading Ca pool %s into /library ." % capoolName)
        caPool = moose.CaConc('/library/'+capoolName)
        poolModel = ionConcElement.find('./{'+self.cml+'}decaying_pool_model')
        caPool.CaBasal = float(poolModel.attrib['resting_conc']) * concfactor
        caPool.Ca_base = float(poolModel.attrib['resting_conc']) * concfactor
        if 'decay_constant' in poolModel.attrib:
            caPool.tau = float(poolModel.attrib['decay_constant']) * Tfactor
        elif 'inv_decay_constant' in poolModel.attrib:
            caPool.tau = 1.0/float(poolModel.attrib['inv_decay_constant']) * Tfactor
        ## Only one of pool_volume_info or fixed_pool_info should be present, but not checking
        volInfo = poolModel.find('./{'+self.cml+'}pool_volume_info')
        if volInfo is not None:
            caPool.thick = float(volInfo.attrib['shell_thickness']) * Lfactor
        fixedPoolInfo = poolModel.find('./{'+self.cml+'}fixed_pool_info')
        if fixedPoolInfo is not None:
            ## Put in phi under the caPool, so that it can
            ## be used instead of thickness to set B (see section 19.2 in Book of Genesis)
            caPool_phi = moose.Mstring(caPool.path+'/phi')
            caPool_phi.value = str( float(fixedPoolInfo.attrib['phi']) * concfactor/Ifactor/Tfactor )
        ## Put a high ceiling and floor for the Ca conc
        #~ self.context.runG('setfield ' + caPool.path + ' ceiling 1e6')
        #~ self.context.runG('setfield ' + caPool.path + ' floor 0.0')

    def make_cml_function(self, element, fn_name, concdep=None):
        fn_type = element.attrib['expr_form']
        if fn_type in ['exponential','sigmoid','exp_linear']:
            self.make_function( fn_name, fn_type, rate=float(element.attrib['rate']),\
                midpoint=float(element.attrib['midpoint']), scale=float(element.attrib['scale'] ) )
        elif fn_type == 'generic':
            ## OOPS! These expressions should be in SI units, since I converted to SI
            ## Ideally I should not convert to SI and have the user consistently use one or the other
            ## Or convert constants in these expressions to SI, only voltage and ca_conc appear!
            ## NO! SEE ABOVE, for calculating values for tables, I use units specified in xml file,
            ## then I convert to SI while writing to MOOSE internal tables.
            expr_string = element.attrib['expr']
            if concdep is None: ca_name = ''                        # no Ca dependence
            else: ca_name = ','+concdep.attrib['variable_name']     # Ca dependence
            expr_string = expr_string.replace( 'alpha', 'self.alpha(v'+ca_name+')')
            expr_string = expr_string.replace( 'beta', 'self.beta(v'+ca_name+')')
            self.make_function( fn_name, fn_type, expr_string=expr_string, concdep=concdep )
        else:
            pu.fatal("Unsupported function type %s "% fn_type)
            sys.exit()

    def make_function(self, fn_name, fn_type, **kwargs):
        """ This dynamically creates a function called fn_name
        If fn_type is exponential, sigmoid or exp_linear,
            **kwargs is a dict having keys rate, midpoint and scale.
        If fin_type is generic, **kwargs is a dict having key expr_string """
        if fn_type == 'exponential':
            def fn(self,v):
                return kwargs['rate']*math.exp((v-kwargs['midpoint'])/kwargs['scale'])
        elif fn_type == 'sigmoid':
            def fn(self,v):
                return kwargs['rate'] / ( 1 + math.exp((v-kwargs['midpoint'])/kwargs['scale']) )
        elif fn_type == 'exp_linear':
            def fn(self,v):
                if v-kwargs['midpoint'] == 0.0: return kwargs['rate']
                else:
                    return kwargs['rate'] * ((v-kwargs['midpoint'])/kwargs['scale']) \
                        / ( 1 - math.exp((kwargs['midpoint']-v)/kwargs['scale']) )
        elif fn_type == 'generic':
            ## python cannot evaluate the ternary operator ?:, so evaluate explicitly
            ## for security purposes eval() is not allowed any __builtins__
            ## but only safe functions in utils.safe_dict and
            ## only the local variables/functions v, self
            allowed_locals = {'self':self}
            allowed_locals.update(utils.safe_dict)
            def fn(self,v,ca=None):
                expr_str = kwargs['expr_string']
                allowed_locals['v'] = v
                allowed_locals['celsius'] = self.temperature
                allowed_locals['temp_adj_'+self.gate_name] = self.q10factor
                for i,parameter in enumerate(self.parameters):
                    allowed_locals[parameter[0]] = self.parameters[i][1]
                if 'concdep' in kwargs:
                    concdep = kwargs['concdep']
                    ## ca should appear as neuroML defined 'variable_name' to eval()
                    if concdep is not None:
                        allowed_locals[concdep.attrib['variable_name']] = ca
                if '?' in expr_str:
                    condition, alternatives = expr_str.split('?',1)
                    alternativeTrue, alternativeFalse = alternatives.split(':',1)
                    if eval(condition,{"__builtins__":None},allowed_locals):
                        val = eval(alternativeTrue,{"__builtins__":None},allowed_locals)
                    else:
                        val = eval(alternativeFalse,{"__builtins__":None},allowed_locals)
                else:
                    val = eval(expr_str,{"__builtins__" : None},allowed_locals)
                if fn_name == 'tau': 
                    return val/self.q10factor
                else: 
                    return val
        fn.__name__ = fn_name
        setattr(self.__class__, fn.__name__, fn)
        return None


def make_new_synapse(syn_name, postcomp, syn_name_full, nml_params):
    ## if channel does not exist in library load it from xml file
    if not moose.exists('/library/'+syn_name):
        cmlR = ChannelML(nml_params)
        model_filename = syn_name+'.xml'
        model_path = utils.find_first_file(model_filename,nml_params['model_dir'])
        if model_path is not None:
            cmlR.readChannelMLFromFile(model_path)
        else:
            raise IOError(
                'For mechanism {0}: files {1} not found under {2}.'.format(
                    syn_name, model_filename, nml_params['model_dir']
                )
            )
    ## deep copies the library SynChan and SynHandler
    ## to instances under postcomp named as <arg3>
    synid = moose.copy(moose.element('/library/'+syn_name),postcomp,syn_name_full)
    #synhandlerid = moose.copy(moose.element('/library/'+syn_name+'/handler'), postcomp,syn_name_full+'/handler') This line was a bug: double handler
    synhandler = moose.element( synid.path + '/handler' )
    syn = moose.SynChan(synid)
    synhandler = moose.element(synid.path + '/handler') # returns SimpleSynHandler or STDPSynHandler

    ## connect the SimpleSynHandler or the STDPSynHandler to the SynChan (double exp)
    moose.connect( synhandler, 'activationOut', syn, 'activation' )
    # mgblock connections if required
    childmgblock = mu.get_child_Mstring(syn,'mgblockStr')
    #### connect the post compartment to the synapse
    if childmgblock.value=='True': # If NMDA synapse based on mgblock, connect to mgblock
        mgblock = moose.Mg_block(syn.path+'/mgblock')
        moose.connect(postcomp,"channel", mgblock, "channel")
    else: # if SynChan or even NMDAChan, connect normally
        moose.connect(postcomp,"channel", syn, "channel")



# -*- coding: utf-8 -*-
# hhfit.py ---

# Description:
# Author:
# Maintainer:
# Created: Tue May 21 16:31:56 2013 (+0530)
# Commentary:
# Functions for fitting common equations for Hodgkin-Huxley type gate
# equations.

import traceback
import warnings
import numpy as np
import logging
logger_ = logging.getLogger('moose.nml2.hhfit')

try:
    import scipy.optimize as _SO
except ImportError:
    raise RuntimeError("To use this feature/module, please install scipy")


def exponential2(x, a, scale, x0, y0=0):
    res = a * np.exp((x - x0) / scale) + y0
    #print('============   Calculating exponential2 for %s, a=%s, scale=%s, x0=%s, y0=%s; = %s'%(x, a, scale, x0, y0, res))
    return res


def exponential(x, a, k, x0, y0=0):
    res = a * np.exp(k * (x - x0)) + y0
    #print('============   Calculating exponential for %s, a=%s, k=%s, x0=%s, y0=%s; = %s'%(x, a, k, x0, y0, res))
    return res


def sigmoid2(x, a, scale, x0, y0=0):
    res = a / (np.exp(-1 * (x - x0) / scale) + 1.0) + y0
    #print('============   Calculating sigmoid for %s, a=%s, scale=%s, x0=%s, y0=%s; = %s'%(x, a, scale, x0, y0, res))
    return res


def sigmoid(x, a, k, x0, y0=0):
    res = a / (np.exp(k * (x - x0)) + 1.0) + y0
    #print('============   Calculating sigmoid for %s, a=%s, k=%s, x0=%s, y0=%s; = %s'%(x, a, k, x0, y0, res))
    return res


def linoid2(x, a, scale, x0, y0=0):
    """The so called linoid function. Called explinear in neuroml."""

    denominator = 1 - np.exp(-1 * (x - x0) / scale)
    # Linoid often includes a zero denominator - we need to fill those
    # points with interpolated values (interpolation is simpler than
    # finding limits).
    ret = (a / scale) * (x - x0) / denominator
    infidx = np.flatnonzero((ret == np.inf) | (ret == -np.inf))
    if len(infidx) > 0:
        for ii in infidx:
            if ii == 0:
                ret[ii] = ret[ii + 1] - (ret[ii + 2] - ret[ii + 1])
            elif ii == len(ret):
                ret[ii] = ret[ii - 1] + (ret[ii - 1] - ret[ii - 2])
            else:
                ret[ii] = (ret[ii + 1] + ret[ii + 2]) * 0.5
    res = ret + y0
    #print('============   Calculating linoid2 for %s, a=%s, scale=%s, x0=%s, y0=%s; res=%s'%(x, a, scale, x0, y0,res))
    return res


def linoid(x, a, k, x0, y0=0):
    """The so called linoid function. Called explinear in neuroml."""

    denominator = np.exp(k * (x - x0)) - 1.0
    # Linoid often includes a zero denominator - we need to fill those
    # points with interpolated values (interpolation is simpler than
    # finding limits).
    ret = a * (x - x0) / denominator
    infidx = np.flatnonzero((ret == np.inf) | (ret == -np.inf))
    if len(infidx) > 0:
        for ii in infidx:
            if ii == 0:
                ret[ii] = ret[ii + 1] - (ret[ii + 2] - ret[ii + 1])
            elif ii == len(ret):
                ret[ii] = ret[ii - 1] + (ret[ii - 1] - ret[ii - 2])
            else:
                ret[ii] = (ret[ii + 1] + ret[ii + 2]) * 0.5
    res = ret + y0
    #print('============   Calculating linoid for %s, a=%s, k=%s, x0=%s, y0=%s; res=%s'%(x, a, k, x0, y0,res))
    return res


def double_exp(x, a, k1, x1, k2, x2, y0=0):
    """For functions of the form:

    a / (exp(k1 * (x - x1)) + exp(k2 * (x - x2)))

    """
    ret = np.zeros(len(x))
    try:
        ret = a / (np.exp(k1 * (x - x1)) + np.exp(k2 * (x - x2))) + y0
    except RuntimeWarning as e:
        logger_.warn(e)
    return ret


# Map from the above functions to corresponding neuroml class
fn_rate_map = {
    exponential: 'HHExpRate',
    sigmoid: 'HHSigmoidRate',
    linoid: 'HHExpLinearRate',
    double_exp: None,
}

# These are default starting parameter values
fn_p0_map = {
    exponential: (1.0, -100, 20e-3, 0.0),
    sigmoid: (1.0, 1.0, 0.0, 0.0),
    linoid: (1.0, 1.0, 0.0, 0.0),
    double_exp: (1e-3, -1.0, 0.0, 1.0, 0.0, 0.0),
}


def randomized_curve_fit(fn, x, y, maxiter=10, best=True):
    """Repeatedly search for a good fit for common gate functions for
    HHtype channels with randomly generated initial parameter
    set. This function first tries with default p0 for fn. If that
    fails to find a good fit, (correlation coeff returned by curve_fit
    being inf is an indication of this), it goes on to generate random
    p0 arrays and try scipy.optimize.curve_fit using this p0 until it
    finds a good fit or the number of iterations reaches maxiter.

    Ideally we should be doing something like stochastic gradient
    descent, but I don't know if that might have performance issue in
    pure python. The random parameterization in the present function
    uses uniformly distributed random numbers within the half-open
    interval [min(x), max(x)). The reason for choosing this: the
    offset used in the exponential parts of Boltzman-type/HH-type
    equations are usually within the domain of x. I also invert the
    second entry (p0[1], because it is always (one of) the scale
    factor(s) and usually 1/v for some v in the domain of x. I have
    not tested the utility of this inversion. Even without this
    inversion, with maxiter=100 this function is successful for the
    test cases.

    Parameters
    ----------
    x: ndarray
    values of the independent variable

    y: ndarray
    sample values of the dependent variable

    maxiter: int
    maximum number of iterations

    best: bool
    if true, repeat curve_fit for maxiter and return the case of least
    squared error.

    Returns
    -------
    The return value of scipy.optimize.curve_fit which succeed, or the
    last call to it if maxiter iterations is reached..

    """
    bad = True
    p0 = fn_p0_map[fn]
    p = None
    p_best = None
    min_err = 1e10  # A large value as placeholder
    for ii in range(maxiter):
        try:
            p = _SO.curve_fit(fn, x, y, p0=p0)
        except (RuntimeError, RuntimeWarning):
            p = None
        # The last entry returned by scipy.optimize.leastsq used by
        # curve_fit is 1, 2, 3 or 4 if it succeeds.
        bad = (p is None) or (p[1] == np.inf).any()
        if not bad:
            if not best:
                return p
            err = sum((y - fn(x, *tuple(p[0])))**2)
            if err < min_err:
                min_err = err
                p_best = p
        p0 = np.random.uniform(low=min(x),
                               high=max(x),
                               size=len(fn_p0_map[fn]))
        if p0[1] != 0.0:
            p0[1] = 1 / p0[1]  # k = 1/v_scale - could help faster convergence
    if p_best is None:
        if p is not None:
            msg = p[-2]
        else:
            msg = ''
        warnings.warn(
            'Maximum iteration %d reached. Could not find a decent fit. %s' %
            (maxiter, msg), RuntimeWarning)
    return p_best


def find_ratefn(x, y, **kwargs):
    """Find the function that fits the rate function best. This will try
    exponential, sigmoid and linoid and return the best fit.

    Needed until NeuroML2 supports tables or MOOSE supports
    functions.

    Parameters
    ----------
    x: 1D array
    independent variable.

    y: 1D array
    function values.

    **kwargs: keyword arguments
    passed to randomized_curve_fit.

    Returns
    -------
    best_fn: function
    the best fit function.

    best_p: tuple
    the optimal parameter values for the best fit function.

    """
    rms_error = 1e10  # arbitrarily setting this
    best_fn = None
    best_p = None
    for fn in fn_rate_map:
        p = randomized_curve_fit(fn, x, y, **kwargs)
        if p is None:
            continue
        popt = p[0]
        error = y - fn(x, *popt)
        erms = np.sqrt(np.mean(error**2))
        # Ideally I want a fuzzy selection criterion here - a
        # preference for fewer parameters, but if the errors are
        # really small then we go for functions with more number of
        # parameters. Some kind of weighted decision would have been
        # nice. I am arbitrarily setting less than 0.1% relative error
        # as a strong argument for taking a longer parameter function
        # as a really better fit. Even with 1%, double exponential
        # betters detected sigmoid for sigmoid curve in test case.
        if erms < rms_error and ((best_p is None) or len(popt) <= len(best_p)
                                 or erms / (max(y) - min(y)) < 0.001):
            rms_error = erms
            best_fn = fn
            best_p = popt
    return (best_fn, best_p)



# -*- coding: utf-8 -*-

"""methods_utils.py:

"""

__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2013, NCBS Bangalore"
__credits__          = ["NCBS Bangalore", "Bhalla Lab"]
__license__          = "GPL"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"


import re

objPathPat = re.compile(r'(\/\w+\[\d+\])+?$')

def idPathToObjPath( idPath ):
    """ Append a [0] if missing from idPath.

    Id-paths do not have [0] at their end. This does not allow one to do
    algebra properly.
    """
    m = objPathPat.match( idPath )
    if m: return idPath
    else:
        return '{}[0]'.format(idPath)


def main():
    p1 = '/cable[0]/comp_[1]/a'
    p2 = '/cab[1]/comp/com'
    p3 = '/cab[1]/p[2]/c[3]'
    p4 = '/ca__b[1]/_p[2]/c[122]'
    for p in [p1, p2, p3, p4]:
        m = objPathPat.match(p)
        if m:
            print(m.group(0))
        else:
            print(("{} is invalid Obj path in moose".format( p )))

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from chan_proto import (ChannelSettings, ZChannelParams,ChannelParams,GateParams)

krev=-87e-3
narev=50e-3
carev=140e-3 #assumes CaExt=2 mM and CaIn=50e-3
ZpowCDI=0

VMIN = -100e-3
VMAX = 50e-3
VDIVS = 3000 #0.5 mV steps
CAMIN=0.01e-3   #10 nM
CAMAX=40e-3  #40 uM, might want to go up to 100 uM with spines
CADIVS=3000 #10 nM steps

CaL12param = ChannelSettings(Xpow=1, Ypow=0, Zpow=0, Erev=0.048, name='CaL12')
qfactCaL = 1
CaL12_X_params = GateParams(A_rate=-880,A_B= -220e3,A_C = -1.0,Avhalf= 4.0003e-3, A_vslope=-7.5e-3,B_rate=-284,B_B=71e3,B_C=-1.0,Bvhalf=-4.0003e-3,B_vslope=5e-3)

SKparam = ChannelSettings(Xpow=0, Ypow=0, Zpow=1, Erev=-.087, name='SKCa')

SK_Z_params = ZChannelParams(Kd = 0.57e-3,
                             power = 5.2,
                             tau = 4.9e-3)


Cal = ChannelParams(channel=CaL12param,X=CaL12_X_params,Y=[],Z=[])
SK = ChannelParams(channel=SKparam,X=[],Y=[],Z=SK_Z_params)



import moose
import numpy as np
import rdesigneur as rd

def test_21_vclamp():
    """Test vclamp.
    """
    rdes = rd.rdesigneur(
        stimList = [['soma', '1', '.', 'vclamp', '-0.065 + (t>0.1 && t<0.2) * 0.02' ]],
        plotList = [
            ['soma', '1', '.', 'Vm', 'Soma membrane potential'],
            ['soma', '1', 'vclamp', 'current', 'Soma holding current'],
        ]
    )
    rdes.buildModel()
    moose.reinit()
    moose.start( 0.3 )
    # rdes.display(block=False)
    data = []
    for t in moose.wildcardFind('/##[TYPE=Table]'):
        data.append(t.vector)
    mean = np.mean(data, axis=1)
    std = np.std(data, axis=1)
    assert np.allclose([-5.83422152e-02, -9.28563233e-09], mean), mean
    assert np.allclose([9.41512562e-03, 2.79081939e-08], std), std
    return data

if __name__ == '__main__':
    test_21_vclamp()



# This example demonstrates periodic synaptic input to a cell.
# Copyright (C) Upinder S. Bhalla NCBS 2018
# Released under the terms of the GNU Public License V3. No warranty.
# Doctest by Dilawar Singh

import moose
import rdesigneur as rd
import numpy as np


def test_periodic_syninput():
    rdes = rd.rdesigneur(
        cellProto = [['somaProto', 'soma', 20e-6, 200e-6]],
        chanProto = [['make_glu()', 'glu']],
        chanDistrib = [['glu', 'soma', 'Gbar', '1' ]],

        # Deliver stimulus to glu synapse on soma, periodically at 50 Hz.
        stimList = [['soma', '0.5', 'glu', 'periodicsyn', '50' ]],
        plotList = [['soma', '1', '.', 'Vm', 'Soma membrane potential']]
    )
    rdes.buildModel()
    moose.reinit()
    moose.start( 0.3 )
    t = moose.wildcardFind('/##[TYPE=Table]')[0].vector
    expected = np.array([-0.04995514162861773, 0.004795008283676097])
    got = np.array([t.mean(), t.std()])
    assert np.allclose(expected, got, rtol=1e-3), (expected, got)
    print('ok')
    return t

if __name__ == '__main__':
    test_periodic_syninput()



# -*- coding: utf-8 -*-
# h10.CNG.moose.py ---
#
# Filename: h10.CNG.moose.py
# Description:
# Author: subha
# Maintainer:
# Created: Sat Apr 30 02:10:09 2016 (-0400)
# Version:
# Last-Updated: Sat Apr 30 03:12:18 2016 (-0400)
#           By: subha
#     Update #: 95
# URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change log:
#
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA.
#
#

# Code:
from __future__ import print_function
import numpy as np
import moose
import rdesigneur as rd

# Reduce dt to get more accurate result from EE solver
dt = 25e-6

model = moose.Neutral('/model')
data = moose.Neutral('/data')
cell = moose.loadModel('h10.CNG.swc', '/model/h10')
RA = 1.0    # Ohm m - nrn 100 Ohm cm
RM = 0.1    # Ohm m2 - nrn 0.001 S/cm2
CM = 0.01   # F/m2  - nrn 1uF/cm2

for comp in moose.wildcardFind('{}/##[ISA=CompartmentBase]'.format(cell.path)):
    sarea = np.pi * comp.diameter * comp.length
    xarea = np.pi * comp.diameter * comp.diameter / 4
    comp.Ra = RA * comp.length / xarea
    comp.Rm = RM / sarea
    comp.Cm = CM * sarea
    comp.initVm = -70e-3
    comp.Em = -70e-3
    print(comp.name, comp.x0, comp.y0, comp.z0, comp.x, comp.y, comp.z)
# Last apical compartment is apical_e_177 between nodes 200 (-3.061,
# 56.39, -3.12) and 204 (-32.910 128.5 -3.12) in the swc file- same as
# 90 in NEURON.

stim = moose.PulseGen('/model/stim')
stim.delay[0] = 20e-3
stim.width[0] = 10e-3
stim.level[0] = 1e-12
stim.delay[1] = 1e9
moose.connect(stim, 'output', moose.element('/model/h10/apical_e_177_0'), 'injectMsg')
tab_v_soma = moose.Table('/data/soma_v')
moose.connect(tab_v_soma, 'requestOut', moose.element('/model/h10/soma'), 'getVm')
tab_v_apical = moose.Table('/data/apical_v')
moose.connect(tab_v_apical, 'requestOut', moose.element('/model/h10/apical_e_177_0'), 'getVm')

import sys

if __name__ == '__main__':
    if len(sys.argv) > 1:
        solver = moose.HSolve('/model/h10/solver')
        solver.target = '/model/h10/soma'
        solver.dt = dt
        outfile = 'h10.CNG.moose.hsolve.txt'
    else:
        outfile = 'h10.CNG.moose.ee.txt'
        for ii in range(20):
            moose.setClock(ii, dt)

    simtime = 100e-3
    moose.reinit()
    moose.start(simtime)
    t = np.arange(len(tab_v_soma.vector)) * simtime / len(tab_v_soma.vector)
    np.savetxt(outfile, np.vstack([t, tab_v_soma.vector, tab_v_apical.vector]).T)



#
# h10.CNG.moose.py ends here



# -*- coding: utf-8 -*-
# mtypes.py ---
#
# Filename: mtypes.py
# Description:
# Author:
# Maintainer:
# Created: Fri Feb  8 11:29:36 2013 (+0530)
# Version:
# Last-Updated: Tue Mar  1 02:52:35 2016 (-0500)
#           By: subha
#     Update #: 182
# URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
# Utility to detect the model type in a file
#
#

# Change log:
#
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA.
#
#

# Code:

from __future__ import print_function
import re
import moose

##!!!!!!!!!!!!!!!!!!!!!!!!!!!!
## !! This was stolen from
## http://eli.thegreenplace.net/2011/10/19/perls-guess-if-file-is-text-or-binary-implemented-in-python/#id2

import sys
PY3 = sys.version_info[0] >= 3

# A function that takes an integer in the 8-bit range and returns
# a single-character byte object in py3 / a single-character string
# in py2.
#
int2byte = (lambda x: bytes((x,))) if PY3 else chr

_text_characters = (
        b''.join(int2byte(i) for i in range(32, 127)) +
        b'\n\r\t\f\b')

def istextfile(fileobj, blocksize=512):
    """ Uses heuristics to guess whether the given file is text or binary,
        by reading a single block of bytes from the file.
        If more than 30% of the chars in the block are non-text, or there
        are NUL ('\x00') bytes in the block, assume this is a binary file.

        - Originally written by Eli Bendersky from the algorithm used
          in Perl:
          http://eli.thegreenplace.net/2011/10/19/perls-guess-if-file-is-text-or-binary-implemented-in-python/#id2
    """
    block = fileobj.read(blocksize)
    if b'\x00' in block:
        # Files with null bytes are binary
        return False
    elif not block:
        # An empty file is considered a valid text file
        return True

    # Use translate's 'deletechars' argument to efficiently remove all
    # occurrences of _text_characters from the block
    nontext = block.translate(None, _text_characters)
    return float(len(nontext)) / len(block) <= 0.30

## Eli Bendersky till here !!
##!!!!!!!!!!!!!!!!!!!!!!!!!!!!


def getType(filename, mode='t'):
    """Returns the type of the model in file `filename`. Returns None
    if type is not known.

    mode: 'b' for binary, 't' for text. Not used currently.

    """
    mtype = None
    msubtype = None
    if mode == 't':
        for typename, typefunc in list(typeChecks.items()):
            if typefunc(filename):
                return typename
    return None

def getSubtype(filename, typename):
    """Returns what subtype of the specified `typename` is the model file.
    None if could not resolve the subtype.
    """
    for subtype in subtypes[typename]:
        subtypeFunc = subtypeChecks['%s/%s' % (typename, subtype)]
        if subtypeFunc(filename):
            return subtype
    return ''

# Dictionary of model description types and functions to detect them.
#
# Fri Feb 8 11:07:41 IST 2013 - as of now we only recognize GENESIS .g
# XML .xml extensions
#
# Fri Feb 22 16:29:43 IST 2013 - added .cspace



isCSPACE = lambda x: x.lower().endswith('.cspace')
isGENESIS = lambda x: x.lower().endswith('.g') or x.endswith('.p')
isXML = lambda x: x.lower().endswith('.xml')

isProto = lambda x: x.lower().endswith('.p')

import xml.dom.minidom as md

def isNeuroML(filename):
    """Check if a model is in neuroml format. An xml document is
    considered a neuroml if the top level element is either
    'networkml', morphml', 'channelml' or 'neuroml'.

    """
    doc = md.parse(filename)
    for child in doc.childNodes:
        print(child.nodeName, child.nodeType == child.ELEMENT_NODE)
        if child.nodeType == child.ELEMENT_NODE and \
                (child.nodeName == 'networkml' or \
                     child.nodeName == 'morphml' or \
                     child.nodeName == 'channelml'or \
                     child.nodeName == 'neuroml'):
            return True
    return  False

def isSBML(filename):
    """Check model in `filename` is in SBML format."""
    doc = md.parse(filename)
    for child in doc.childNodes:
        if child.nodeType == child.ELEMENT_NODE and \
                child.nodeName == 'sbml':
            return True
    return  False

def isKKIT(filename):
    """Check if `filename` is a GENESIS/KINETIKIT file.

    """
    pattern = re.compile('include\s+kkit') # KKIT files must have "include kkit" statement somewhere
    with open(filename, 'r') as infile:
        while True:
            sentence = ''
            contd = False # Flags if we are inside a multi-line entry
            line = infile.readline()
            if not line: # End of file
                return False
            line = line.strip()
            # print 'read:', line
            if line.find('//') == 0: # skip c++ style comment lines
                # print 'c++ comment'
                continue
            # Skip C style multi-line comments
            comment_start = line.find('/*')
            if comment_start >= 0:
                sentence = line[:comment_start]
                # print 'cstart', comment_start, sentence
            while comment_start >= 0 and line:
                # print '#', line
                comment_end = line.find('*/')
                if comment_end >= 0:
                    comment_start = -1;
                    line = line[comment_end+2:] # keep the rest of the line
                    break
                line = infile.readline()
                if line:
                    line = line.strip()
            # Keep extending the sentence with next line if current
            # line ends with continuation token '\'
            if line:
                contd = line.endswith('\\')
            while line and contd:
                sentence += ' ' + line[:-1]
                line = infile.readline()
                if line:
                    line = line.strip()
                    contd = line.endswith('\\')
            # if contd turned false, the last line came out of the
            # while loop unprocessed
            if line:
                sentence += ' ' + line
            # print '>', sentence
            if re.search(pattern, sentence):
                return True
    return False

# Mapping model types to functions to check them.  These functions
# should take the filename as the single argument and return True or
# False. Define new functions for new model type checks and add them
# here.
typeChecks = {
    'cspace': isCSPACE,
    'genesis': isGENESIS,
    'xml': isXML,
}

# These are "type/subtype": function maps for checking model subtype
# New model types with subtypes should be added here with
# corresponding checker function.
subtypeChecks = {
    'genesis/kkit': isKKIT,
    'genesis/proto': isProto,
    'xml/neuroml': isNeuroML,
    'xml/sbml': isSBML,
}

# Mapping types to list of subtypes.
subtypes = {
    'cspace': [],
    'genesis': ['kkit', 'proto'],
    'xml': ['neuroml', 'sbml'],
}



#
# mtypes.py ends here



# -*- coding: utf-8 -*-
"""test_socket_streamer.py:

    MOOSE must create a socket server on PORT 31616 (default)
    or setup moose.SocketStreamer port to appropriate port number.

    Client can read data from this socket.
"""

from __future__ import print_function

__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2016, Dilawar Singh"
__credits__          = ["NCBS Bangalore"]
__license__          = "GNU GPL"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import os
import sys
sys.path.append(os.path.dirname(__file__))
import time
import socket
import numpy as np
import multiprocessing as mp
import moose
import moose.utils as mu
import json
import models
from collections import defaultdict

print( '[INFO] Using moose form %s' % moose.__file__ )

sockFile_ = '/tmp/MOOSE'

def socket_client(done, q):
    # This function waits for socket to be available.
    global sockFile_
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    address = sockFile_
    while not os.path.exists(address):
        if done.value == 1:
            return
        continue
    s.connect( address )
    print( 'Py: Connected with socket. %s' % sockFile_ )

    # This is client reponsibility to read the data.
    print( 'Py: Fetching...' )
    data = b''
    s.settimeout(0.01)
    while True:
        try:
            data += s.recv(64)
        except socket.timeout:
            print('x', end = '' )

        if done.value == 1:
            print( 'Simulation is over' )
            break
    s.close()
    if not data:
        print("No data streamed")
        done.value = 1
        q.put({})
        return
    res = mu.decode_data(data)
    q.put(res)

def test():
    q = mp.Queue()
    done = mp.Value( 'd', 0.0)
    client = mp.Process(target=socket_client, args=(done, q))
    client.start()
    print( '[INFO] Socket client is running now' )
    time.sleep(0.1)

    # Now create a streamer and use it to write to a stream
    os.environ['MOOSE_STREAMER_ADDRESS'] = 'file://%s' % sockFile_
    models.simple_model_a()

    # Give some time for socket client to make connection.
    moose.reinit()
    moose.start(50)
    time.sleep(0.1)
    done.value = 1

    res = q.get()
    if not res:
        raise RuntimeWarning( 'Nothing was streamed')
    for k in res:
        aWithTime = res[k]
        a = aWithTime[1::2]
        b = moose.element(k).vector
        print(k, len(a), len(b))
        assert (a == b).all()

    # sleep for some time so data can be read.
    client.join()

def main( ):
    test()
    print( '[INFO] All tests passed' )

if __name__ == '__main__':
    main()



import moose
print( 'Using moose from %s' % moose.__file__)

modelRoot = moose.Neutral('/model')
compt = moose.CubeMesh('/model/kinetics')
p1 = moose.BufPool('/model/kinetics/Ca')
function = moose.Function('/model/kinetics/Ca/func')
#function out to pool
moose.connect( function, 'valueOut', p1 ,'setConc' )
#input pools to be connected to fuction
S1 = moose.BufPool( '/model/kinetics/s1' )
S2 = moose.BufPool( '/model/kinetics/s2' )

initiexpr = function.expr
#S1 connected to function
numVariables = function.numVars
expr = ""
expr = (function.expr+'+'+'x'+str(numVariables))
expr = expr.lstrip("0 +")
expr = expr.replace(" ","")
function.expr = expr
s1msg = moose.connect( S1, 'nOut', function.x[numVariables], 'input' )
#S2 connected to function
numVariables = function.numVars
expr = ""
expr = (function.expr+'+'+'x'+str(numVariables))
expr = expr.lstrip("0 +")
expr = expr.replace(" ","")
function.expr = expr
s2msg = moose.connect( S2, 'nOut', function.x[numVariables], 'input' )

compts = moose.wildcardFind(modelRoot.path+'/##[ISA=ChemCompt]')
solver = 'gsl'
moose.mooseAddChemSolver('/model',solver)
moose.mooseDeleteChemSolver('/model')

#moose.delete(s1msg)
#moose.delete(s2msg)
#deleting the function
#moose.delete(function) #deleting this will not cause seg fault
print(function.expr) #function's expr still has the expression which is causing the problem
